"""Stored command arguments for plugins."""

from __future__ import annotations

import shlex
import typing as t

from meltano.core.behavior.canonical import Canonical
from meltano.core.container.container_spec import ContainerSpec
from meltano.core.error import Error
from meltano.core.utils import expand_env_vars

TCommand = t.TypeVar("TCommand")


class UndefinedEnvVarError(Error):
    """An environment variable is used as a command argument but is not set."""

    def __init__(self, command_name, var) -> None:  # noqa: ANN001
        """Initialize UndefinedEnvVarError.

        Args:
            command_name: Plugin command name.
            var: Environment variable name.
        """
        super().__init__(
            f"Command '{command_name}' referenced unset environment variable "
            f"'{var}' in an argument. Set the environment variable or update "
            "the command definition.",
        )


class Command(Canonical):
    """This class represents stored command arguments for plugins."""

    def __init__(
        self,
        args: str,
        description: str | None = None,
        executable: str | None = None,
        container_spec: dict | None = None,
    ):
        """Initialize a Command.

        Args:
            args: Command arguments.
            description: Command description.
            executable: Optional command executable.
            container_spec: Container specification for this command.
        """
        super().__init__(
            args=args,
            description=description,
            executable=executable,
        )
        if container_spec is not None:
            self.container_spec = ContainerSpec(**container_spec)
        else:
            self.container_spec = None

    def expanded_args(self, name, env):  # noqa: ANN001, ANN201
        """Replace any env var arguments with their values.

        Args:
            name: Command name.
            env: Mapping of environment variables to expand the command.

        Returns:
            List of expanded command parts.

        Raises:
            UndefinedEnvVarError: if an env var argument is not set
        """
        expanded = []
        for arg in shlex.split(self.args):
            if value := expand_env_vars(arg, env):
                expanded.append(value)

            else:
                raise UndefinedEnvVarError(name, arg)

        return expanded

    @classmethod
    def as_canonical(cls, target):  # noqa: ANN001, ANN206
        """Serialize the target command.

        Args:
            target: Target object type.

        Returns:
            Python object.
        """
        canonical = super().as_canonical(target)
        # if there are only args, flatten the object
        # to the short form
        if "args" in canonical and len(canonical) == 1:
            return canonical["args"]

        return canonical

    @classmethod
    def parse(cls, obj):  # noqa: ANN001, ANN206
        """Deserialize data into a Command.

        Args:
            obj: Raw Python object.

        Returns:
            Command instance.
        """
        if isinstance(obj, str):
            # allow setting the arguments as the value
            # without a description
            return Command(args=obj)

        return super().parse(obj)

    @classmethod
    def parse_all(cls: type[TCommand], obj: dict | None) -> dict[str, TCommand]:
        """Deserialize commands data into a dict of Commands.

        Args:
            obj: Raw Python object.

        Returns:
            Mapping of command names to instances.
        """
        if obj is not None:
            return {name: Command.parse(cmd) for name, cmd in obj.items()}

        return {}



from __future__ import annotations

import pytest


@pytest.fixture(scope="session")
def engine_uri(tmp_path_factory, worker_id: str) -> str:
    database_path = (
        tmp_path_factory.mktemp("fixture_db_sqlite_engine_uri")
        / f"pytest_meltano_{worker_id}.db"
    )
    return f"sqlite:///{database_path.as_posix()}"



"""Storage managers for job state."""

from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod, abstractproperty

if t.TYPE_CHECKING:
    from meltano.core.job_state import JobState


class UnsupportedStateBackendURIError(Exception):
    """Provided state backend URI is not supported."""


class MissingStateBackendSettingsError(Exception):
    """Required setting values for a configured State Backend are missing."""


class StateIDLockedError(Exception):
    """A job attempted to acquire a lock on an already-locked state ID."""


class StateStoreManager(ABC):
    """Base state store manager."""

    def __init__(self, **kwargs) -> None:  # noqa: ANN003, B027
        """Initialize state store manager.

        Args:
            kwargs: additional keyword arguments
        """

    @abstractproperty
    def label(self) -> str:
        """Get the label for the StateStoreManager."""
        ...

    @abstractmethod
    def set(self, state: JobState):  # noqa: ANN201
        """Set the job state for the given state_id.

        Args:
            state: the state to set.

        Raises:
            NotImplementedError: always, this is an abstract method
        """
        ...

    @abstractmethod
    def get(self, state_id) -> JobState | None:  # noqa: ANN001
        """Get the job state for the given state_id.

        Args:
            state_id: the name of the job to get state for.

        Raises:
            NotImplementedError: always, this is an abstract method
        """
        ...

    @abstractmethod
    def clear(self, state_id):  # noqa: ANN001, ANN201
        """Clear state for the given state_id.

        Args:
            state_id: the state_id to clear state for
        """
        ...

    @abstractmethod
    def get_state_ids(self, pattern=None):  # noqa: ANN001, ANN201
        """Get all state_ids available in this state store manager.

        Args:
            pattern: glob-style pattern to filter by
        """
        ...

    @abstractmethod
    def acquire_lock(self, state_id):  # noqa: ANN001, ANN201
        """Acquire a naive lock for the given job's state.

        Args:
            state_id: the state_id to lock
        """
        ...



from __future__ import annotations

import errno
import json
import os
import shutil
import typing as t

import mock
import pytest
import yaml
from sqlalchemy.exc import OperationalError

from meltano.core.plugin_remove_service import PluginRemoveService

if t.TYPE_CHECKING:
    from meltano.core.plugin_location_remove import PluginLocationRemoveManager


class TestPluginRemoveService:
    @pytest.fixture()
    def subject(self, project):
        return PluginRemoveService(project)

    @pytest.fixture()
    def add(self, subject: PluginRemoveService) -> t.Generator[None, None, None]:
        with subject.project.meltanofile.open("r") as meltano_yml:
            original = yaml.safe_load(meltano_yml)

        with subject.project.meltanofile.open("w") as meltano_yml:
            meltano_yml.write(
                yaml.dump(
                    {
                        "plugins": {
                            "extractors": [
                                {
                                    "name": "tap-gitlab",
                                    "variant": "meltanolabs",
                                    "pip_url": "git+https://github.com/MeltanoLabs/tap-gitlab.git",
                                },
                            ],
                            "loaders": [
                                {
                                    "name": "target-csv",
                                    "variant": "meltanolabs",
                                    "pip_url": "git+https://github.com/MeltanoLabs/target-csv.git",
                                },
                            ],
                        },
                    },
                ),
            )

        yield

        with subject.project.meltanofile.open("w") as meltano_yml:
            meltano_yml.write(yaml.dump(original))

    @pytest.fixture()
    def no_plugins(self, subject: PluginRemoveService) -> t.Generator[None, None, None]:
        with subject.project.meltanofile.open("r") as meltano_yml:
            original = yaml.safe_load(meltano_yml)

        with subject.project.meltanofile.open("w") as meltano_yml:
            meltano_yml.write(yaml.dump({"plugins": {}}))

        yield

        with subject.project.meltanofile.open("w") as meltano_yml:
            meltano_yml.write(yaml.dump(original))

    @pytest.fixture()
    def install(self, subject: PluginRemoveService) -> t.Generator[None, None, None]:
        tap_gitlab_installation = subject.project.meltano_dir().joinpath(
            "extractors",
            "tap-gitlab",
        )
        target_csv_installation = subject.project.meltano_dir().joinpath(
            "loaders",
            "target-csv",
        )
        tap_gitlab_installation.mkdir(parents=True, exist_ok=True)
        target_csv_installation.mkdir(parents=True, exist_ok=True)
        yield
        shutil.rmtree(tap_gitlab_installation, ignore_errors=True)
        shutil.rmtree(target_csv_installation, ignore_errors=True)

    @pytest.fixture()
    def lock(self, subject: PluginRemoveService) -> t.Generator[None, None, None]:
        tap_gitlab_lockfile = subject.project.plugin_lock_path(
            "extractors",
            "tap-gitlab",
            variant_name="meltanolabs",
        )
        target_csv_lockfile = subject.project.plugin_lock_path(
            "loaders",
            "target-csv",
            variant_name="meltanolabs",
        )
        with tap_gitlab_lockfile.open("w") as f:
            json.dump(
                {
                    "plugin_type": "extractors",
                    "name": "tap-gitlab",
                    "namespace": "tap_gitlab",
                    "variant": "meltanolabs",
                },
                f,
            )

        with target_csv_lockfile.open("w") as f:
            json.dump(
                {
                    "plugin_type": "loaders",
                    "name": "target-csv",
                    "namespace": "target_csv",
                    "variant": "meltanolabs",
                },
                f,
            )
        yield
        tap_gitlab_lockfile.unlink(missing_ok=True)
        target_csv_lockfile.unlink(missing_ok=True)

    def test_default_init_should_not_fail(self, subject) -> None:
        assert subject

    @pytest.mark.usefixtures("add", "install", "lock")
    def test_remove(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()
        plugins = list(subject.project.plugins.plugins())
        removed_plugins, total_plugins = subject.remove_plugins(plugins)

        assert removed_plugins == total_plugins

        for plugin in plugins:
            # check removed from meltano.yml
            with subject.project.meltanofile.open() as meltanofile:
                meltano_yml = yaml.safe_load(meltanofile)

                with pytest.raises(KeyError):
                    meltano_yml[plugin.type, plugin.name]

            # check removed installation
            path = subject.project.meltano_dir().joinpath(plugin.type, plugin.name)
            assert not path.exists()

            # check removed lock files
            lock_file_paths = list(
                subject.project.root_plugins_dir(plugin.type).glob(
                    f"{plugin.name}*.lock",
                ),
            )
            assert all(not path.exists() for path in lock_file_paths)

    @pytest.mark.usefixtures("no_plugins")
    def test_remove_not_added_or_installed(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()
        plugins = list(subject.project.plugins.plugins())
        removed_plugins, total_plugins = subject.remove_plugins(plugins)

        assert removed_plugins == total_plugins == 0

    @pytest.mark.usefixtures("add", "install", "lock")
    def test_remove_db_error(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()
        plugins = list(subject.project.plugins.plugins())

        assert plugins

        errors = []

        def _collect_error(manager: PluginLocationRemoveManager) -> None:
            errors.append(manager.message)

        with mock.patch(
            "meltano.core.plugin_location_remove.PluginSettingsService.reset",
        ) as reset:
            reset.side_effect = OperationalError(
                "DELETE FROM plugin_settings WHERE plugin_settings.namespace = ?",
                ("extractors.tap-csv.default"),
                "attempt to write a readonly database",
            )
            removed_plugins, _ = subject.remove_plugins(
                plugins,
                removal_manager_status_cb=_collect_error,
            )

        assert removed_plugins == 0
        assert errors.count("attempt to write a readonly database") == len(plugins)

    @pytest.mark.usefixtures("add", "install", "lock")
    def test_remove_meltano_yml_error(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()

        def raise_permissionerror(filename) -> t.NoReturn:
            raise OSError(errno.EACCES, os.strerror(errno.ENOENT), filename)

        plugins = list(subject.project.plugins.plugins())
        with mock.patch.object(
            subject.project.plugins,
            "remove_from_file",
            side_effect=raise_permissionerror,
        ):
            removed_plugins, total_plugins = subject.remove_plugins(plugins)

        assert removed_plugins == 0

    @pytest.mark.usefixtures("add", "install", "lock")
    def test_remove_installation_error(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()

        def raise_permissionerror(filename) -> t.NoReturn:
            raise OSError(errno.EACCES, os.strerror(errno.ENOENT), filename)

        plugins = list(subject.project.plugins.plugins())

        with mock.patch("meltano.core.plugin_location_remove.shutil.rmtree") as rmtree:
            rmtree.side_effect = raise_permissionerror
            removed_plugins, total_plugins = subject.remove_plugins(plugins)

        assert removed_plugins == 0

    @pytest.mark.usefixtures("add", "install")
    def test_remove_lockfile_not_found(self, subject: PluginRemoveService) -> None:
        subject.project.refresh()
        plugins = list(subject.project.plugins.plugins(ensure_parent=False))
        removed_plugins, _ = subject.remove_plugins(plugins)

        assert removed_plugins == 0



from __future__ import annotations

import typing as t

from meltano.core.block.plugin_command import plugin_command_invoker
from meltano.core.tracking.contexts import PluginsTrackingContext
from meltano.core.tracking.schemas import PluginsContextSchema
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project


class TestPluginsTrackingContext:
    def test_plugins_tracking_context_from_block(
        self,
        project: Project,
        dbt: ProjectPlugin,
    ) -> None:
        plugin_ctx = PluginsTrackingContext.from_block(
            plugin_command_invoker(
                dbt,
                project,
                command="test",
            ),
        )
        assert plugin_ctx.schema == PluginsContextSchema.url
        assert len(plugin_ctx.data.get("plugins")) == 1
        plugin_dict = plugin_ctx.data.get("plugins")[0]
        self.assert_plugin_attributes(plugin_dict, dbt)
        assert plugin_dict.get("command") == "test"

    def test_plugins_tracking_context(
        self,
        tap: ProjectPlugin,
        dbt: ProjectPlugin,
    ) -> None:
        plugin_ctx = PluginsTrackingContext([(tap, None), (dbt, "test")])
        assert plugin_ctx.schema == PluginsContextSchema.url
        assert len(plugin_ctx.data.get("plugins")) == 2
        for plugin_dict in plugin_ctx.data.get("plugins"):
            if plugin_dict.get("category") == "extractors":
                self.assert_plugin_attributes(plugin_dict, tap)
                assert plugin_dict.get("command") is None
            elif plugin_dict.get("category") == "transformers":
                self.assert_plugin_attributes(plugin_dict, dbt)
                assert plugin_dict.get("command") == "test"

        # Verify that passing a None object results in an empty plugin context.
        plugin_ctx = PluginsTrackingContext([(None, None)])
        assert plugin_ctx.data.get("plugins") == [{}]

        # Verify that passing a plugin with no parent does not result in an
        # error. Most likely this is a plugin that is not installed and is
        # being removed or somehow referenced.
        tap.parent = None
        plugin_ctx = PluginsTrackingContext([(tap, None)])
        assert len(plugin_ctx.data.get("plugins")) == 1
        plugin_with_no_parent = plugin_ctx.data.get("plugins")[0]
        assert plugin_with_no_parent.get("name_hash") == hash_sha256(tap.name)
        assert not plugin_with_no_parent.get("parent_name_hash")

    @staticmethod
    def assert_plugin_attributes(
        plugin_dict: dict[str, t.Any],
        plugin: ProjectPlugin,
    ) -> None:
        for dict_key, plugin_key in (
            ("name_hash", "name"),
            ("namespace_hash", "namespace"),
            ("executable_hash", "executable"),
            ("variant_name_hash", "variant"),
            ("pip_url_hash", "pip_url"),
        ):
            assert plugin_dict.get(dict_key) == hash_sha256(getattr(plugin, plugin_key))
        assert plugin_dict.get("parent_name_hash") == hash_sha256(plugin.parent.name)



from __future__ import annotations

import asyncio
import tempfile

import mock
import pytest
import structlog
from mock import AsyncMock, Mock
from structlog.testing import capture_logs

from meltano.core.block.singer import SingerBlock
from meltano.core.job import Job
from meltano.core.logging import OutputLogger


class TestSingerBlocks:
    @pytest.fixture()
    def log(self, tmp_path):
        return tempfile.NamedTemporaryFile(mode="w+", dir=tmp_path)

    @pytest.fixture()
    def elt_context(
        self,
        project,  # noqa: ARG002
        session,
        tap,
        target,
        elt_context_builder,
    ):
        job = Job(job_name="pytest_test_runner")

        return (
            elt_context_builder.with_session(session)
            .with_extractor(tap.name)
            .with_job(job)
            .with_loader(target.name)
            .context()
        )

    @pytest.fixture()
    def process_mock_factory(self):
        def _factory(name):
            process_mock = Mock()
            process_mock.name = name
            process_mock.wait = AsyncMock(return_value=0)
            return process_mock

        return _factory

    @pytest.fixture()
    def mock_tap_plugin_invoker(self, process_mock_factory, tap):
        stdout_lines = (b"out1\n", b"out2\n", b"out3\n")
        stderr_lines = (b"err1\n", b"err2\n", b"err3\n")

        tap_process = process_mock_factory(tap)
        tap_process.stdout = mock.MagicMock()
        tap_process.stdout.at_eof.side_effect = (False, False, False, True)
        tap_process.stdout.readline = AsyncMock(side_effect=stdout_lines)
        tap_process.stderr = mock.MagicMock()
        tap_process.stderr.at_eof.side_effect = (False, False, False, True)
        tap_process.stderr.readline = AsyncMock(side_effect=stderr_lines)

        tap_process.stdin = mock.MagicMock()
        tap_process.stdin.wait_closed = AsyncMock(return_value=True)

        tap_process.wait = AsyncMock(return_value=0)

        invoker = Mock()
        invoker.invoke_async = AsyncMock(return_value=tap_process)
        invoker.plugin = tap
        invoker.cleanup = AsyncMock()
        return invoker

    @pytest.fixture()
    def mock_target_plugin_invoker(self, process_mock_factory, target):
        target_process = process_mock_factory(target)
        target_process.stdin = mock.MagicMock()
        target_process.stdin.wait_closed = AsyncMock(return_value=True)

        invoker = Mock()
        invoker.invoke_async = AsyncMock(return_value=target_process)
        invoker.plugin = target
        invoker.cleanup = AsyncMock()
        return invoker

    @pytest.mark.asyncio()
    async def test_singer_block_start(
        self,
        elt_context,
        mock_tap_plugin_invoker,
        mock_target_plugin_invoker,
    ) -> None:
        block = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_tap_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        assert block.producer
        await block.start()
        assert mock_tap_plugin_invoker.invoke_async.called
        assert mock_tap_plugin_invoker.invoke_async.call_args[1]["stdin"] is None
        assert (
            mock_tap_plugin_invoker.invoke_async.call_args[1]["stdout"]
            == asyncio.subprocess.PIPE
        )
        assert (
            mock_tap_plugin_invoker.invoke_async.call_args[1]["stderr"]
            == asyncio.subprocess.PIPE
        )

        block = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_target_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        assert block.consumer
        await block.start()
        assert mock_target_plugin_invoker.invoke_async.called
        assert (
            mock_target_plugin_invoker.invoke_async.call_args[1]["stdin"]
            == asyncio.subprocess.PIPE
        )
        assert (
            mock_target_plugin_invoker.invoke_async.call_args[1]["stdout"]
            == asyncio.subprocess.PIPE
        )
        assert (
            mock_target_plugin_invoker.invoke_async.call_args[1]["stderr"]
            == asyncio.subprocess.PIPE
        )

    @pytest.mark.asyncio()
    async def test_singer_block_stop(
        self,
        elt_context,
        mock_target_plugin_invoker,
    ) -> None:
        block = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_target_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        await block.start()

        await block.stop(kill=True)
        assert block.process_handle.kill.called
        assert block.invoker.cleanup.called

        block = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_target_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        await block.start()

        await block.stop(kill=False)
        assert block.process_handle.terminate.called
        assert block.invoker.cleanup.called

    @pytest.mark.asyncio()
    async def test_singer_block_io(
        self,
        elt_context,
        mock_tap_plugin_invoker,
        log,
    ) -> None:
        producer = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_tap_plugin_invoker,
            plugin_args={"foo": "bar"},
        )

        mock_tap_plugin_invoker.output_handlers = []  # we're capturing logs below directly  # noqa: E501

        output_log = OutputLogger(log)
        log = structlog.getLogger("test")
        producer.stdout_link(output_log.out("stdout", log))
        producer.stderr_link(output_log.out("stderr", log))

        # This test is a great proxy for general io tests
        # if you link the output logger, you can use structlog's capture method
        # to capture the output and check output was actually consumed AND
        # linked correctly.
        with capture_logs() as cap_logs:
            await producer.start()

            io_futures = [producer.proxy_stdout(), producer.proxy_stderr()]

            done, _ = await asyncio.wait(
                io_futures,
                return_when=asyncio.FIRST_COMPLETED,
            )

            expected_lines = [
                {"name": "stdout", "event": "out1", "log_level": "info"},
                {"name": "stdout", "event": "out2", "log_level": "info"},
                {"name": "stdout", "event": "out3", "log_level": "info"},
                {"name": "stderr", "event": "err1", "log_level": "info"},
                {"name": "stderr", "event": "err2", "log_level": "info"},
                {"name": "stderr", "event": "err3", "log_level": "info"},
            ]

            assert cap_logs == expected_lines

    @pytest.mark.asyncio()
    async def test_singer_block_close_stdin(
        self,
        elt_context,
        mock_tap_plugin_invoker,
        mock_target_plugin_invoker,
    ) -> None:
        producer = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_tap_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        assert producer.producer

        await producer.start()
        await producer.close_stdin()
        assert producer.process_handle.stdin.wait_closed.call_count == 0

        consumer = SingerBlock(
            block_ctx=elt_context,
            project=elt_context.project,
            plugin_invoker=mock_target_plugin_invoker,
            plugin_args={"foo": "bar"},
        )
        assert consumer.consumer

        await consumer.start()
        await consumer.close_stdin()
        assert consumer.process_handle.stdin.wait_closed.call_count == 1



"""Base Error classes."""

from __future__ import annotations

import typing as t
from enum import Enum

if t.TYPE_CHECKING:
    from asyncio.streams import StreamReader
    from asyncio.subprocess import Process

    from meltano.core.project import Project


class ExitCode(int, Enum):  # noqa: D101
    OK = 0
    FAIL = 1
    NO_RETRY = 2


class Error(Exception):
    """Base exception for ELT errors."""

    def exit_code(self):  # noqa: ANN201, D102
        return ExitCode.FAIL


class MeltanoError(Error):
    """Base class for all user-facing errors."""

    def __init__(
        self,
        reason: str | Exception | None = None,
        instruction: str | None = None,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """Initialize a MeltanoError.

        Args:
            reason: A short explanation of the error.
            instruction: A short instruction on how to fix the error.
            args: Additional arguments to pass to the base exception class.
            kwargs: Keyword arguments to pass to the base exception class.
        """
        self.reason = reason
        self.instruction = instruction
        super().__init__(reason, instruction, *args, **kwargs)

    def __str__(self) -> str:
        """Return a string representation of the error.

        Returns:
            A string representation of the error.
        """
        return (
            f"{self.reason}. {self.instruction}."
            if self.instruction
            else f"{self.reason}."
        )


class ExtractError(Error):
    """Error in the extraction process, like API errors."""

    def exit_code(self):  # noqa: ANN201, D102
        return ExitCode.NO_RETRY


class AsyncSubprocessError(Exception):
    """Happens when an async subprocess exits with a resultcode != 0."""

    def __init__(
        self,
        message: str,
        process: Process,
        stderr: str | None = None,
    ):
        """Initialize AsyncSubprocessError."""
        self.process = process
        self._stderr: str | StreamReader | None = stderr or process.stderr
        super().__init__(message)

    @property
    async def stderr(self) -> str | None:
        """Return the output of the process to stderr."""
        if not self._stderr:
            return None
        if not isinstance(self._stderr, str):
            stream = await self._stderr.read()
            self._stderr = stream.decode("utf-8")

        return self._stderr


class PluginInstallError(Exception):
    """Exception for when a plugin fails to install."""


class PluginInstallWarning(Exception):
    """Exception for when a plugin optional optional step fails to install."""


class EmptyMeltanoFileException(MeltanoError):
    """Exception for empty meltano.yml file."""

    def __init__(self) -> None:
        """Instantiate the error."""
        reason = "Your meltano.yml file is empty"
        instruction = "Please update your meltano file with a valid configuration"
        super().__init__(reason, instruction)


class MeltanoConfigurationError(MeltanoError):
    """Exception for when Meltano is improperly configured."""


class ProjectNotFound(Error):
    """A Project is instantiated outside of a meltano project structure."""

    def __init__(self, project: Project):
        """Instantiate the error.

        Args:
            project: the name of the project which cannot be found
        """
        super().__init__(
            f"Cannot find `{project.meltanofile}`. Are you in a meltano project?",
        )


class ProjectReadonly(Error):
    """Attempting to update a readonly project."""

    def __init__(self) -> None:
        """Instantiate the error."""
        super().__init__("This Meltano project is deployed as read-only")



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pytest

from meltano.core.job import Job
from meltano.core.job.stale_job_failer import fail_stale_jobs


class TestStaleJobFailer:
    @pytest.fixture()
    def live_job(self, session):
        job = Job(job_name="test")
        job.start()
        job.save(session)

        return job

    @pytest.fixture()
    def stale_job(self, session):
        job = Job(job_name="test")
        job.start()
        job.last_heartbeat_at = datetime.now(timezone.utc) - timedelta(minutes=10)
        job.save(session)

        return job

    @pytest.fixture()
    def other_stale_job(self, session):
        job = Job(job_name="other")
        job.start()
        job.last_heartbeat_at = datetime.now(timezone.utc) - timedelta(minutes=10)
        job.save(session)

        return job

    @pytest.fixture()
    def complete_job(self, session):
        job = Job(job_name="other")
        job.start()
        job.success()
        job.save(session)

        return job

    def test_fail_stale_jobs(
        self,
        live_job,
        stale_job,
        other_stale_job,
        complete_job,
        session,
    ) -> None:
        assert stale_job.is_stale()
        assert other_stale_job.is_stale()

        fail_stale_jobs(session)

        session.refresh(live_job)
        session.refresh(stale_job)
        session.refresh(other_stale_job)
        session.refresh(complete_job)

        # Leaves non-stale jobs alone
        assert live_job.is_running()
        assert complete_job.is_complete()

        # Marks all stale jobs as failed
        assert stale_job.has_error()
        assert not stale_job.is_stale()

        assert other_stale_job.has_error()
        assert not other_stale_job.is_stale()

    def test_fail_stale_jobs_with_state_id(
        self,
        live_job,
        stale_job,
        other_stale_job,
        complete_job,
        session,
    ) -> None:
        assert stale_job.is_stale()
        assert other_stale_job.is_stale()

        fail_stale_jobs(session, state_id=stale_job.job_name)

        session.refresh(live_job)
        session.refresh(stale_job)
        session.refresh(other_stale_job)
        session.refresh(complete_job)

        # Leaves non-stale jobs alone
        assert live_job.is_running()
        assert complete_job.is_complete()

        # Marks stale jobs with the state ID as failed
        assert stale_job.has_error()
        assert not stale_job.is_stale()

        # Leaves stale jobs with a different state ID alone
        assert other_stale_job.is_stale()



"""Create meltano.api tables.

Revision ID: 6ef30ab7b8e5
Revises: b4c05e463b53
Create Date: 2019-07-23 16:09:20.192447

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "6ef30ab7b8e5"
down_revision = "b4c05e463b53"
branch_labels = None
depends_on = None


def upgrade() -> None:
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)
    max_string_length = max_string_length_for_dialect(dialect_name)

    op.create_table(
        "user",
        sa.Column("id", sa.Integer, nullable=False),
        sa.Column("username", sa.String(255)),
        sa.Column("email", sa.String(255)),
        sa.Column("password", sa.String(255)),
        sa.Column("active", sa.Boolean),
        sa.Column("confirmed_at", datetime_type, nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email"),
    )
    op.create_index("ix_user_username", "user", ["username"], unique=1)

    op.create_table(
        "role",
        sa.Column("id", sa.Integer, nullable=False),
        sa.Column("name", sa.String(80)),
        sa.Column("description", sa.String(255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )

    op.create_table(
        "roles_users",
        sa.Column("id", sa.Integer, nullable=False),
        sa.Column("user_id", sa.Integer),
        sa.Column("role_id", sa.Integer),
        sa.ForeignKeyConstraint(["role_id"], ["role.id"]),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"]),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "role_permissions",
        sa.Column("id", sa.Integer, nullable=False),
        sa.Column("role_id", sa.Integer),
        sa.Column("type", sa.String(max_string_length)),
        sa.Column("context", sa.String(max_string_length), nullable=True),
        sa.ForeignKeyConstraint(["role_id"], ["role.id"]),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "oauth",
        sa.Column("id", sa.Integer, nullable=False),
        sa.Column("user_id", sa.Integer),
        sa.Column("provider_id", sa.String(255)),
        sa.Column("provider_user_id", sa.Integer, nullable=True),
        sa.Column("access_token", sa.String(255)),
        sa.Column("created_at", datetime_type, nullable=True),
        sa.Column("id_token", sa.String(max_string_length)),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"]),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade() -> None:
    op.drop_table("roles_users")
    op.drop_table("role_permissions")
    op.drop_table("role")
    op.drop_index("ix_user_username", table_name="user")
    op.drop_table("user")
    op.drop_table("oauth")



from __future__ import annotations

from typing import NoReturn  # noqa: ICN003

import mock
import pytest

from meltano.core.behavior.hookable import HookObject, hook


class Hooked(HookObject):
    def __init__(self) -> None:
        self._calls = []

    def call(self, hook_name) -> None:
        self._calls.append(hook_name)

    @property
    def calls(self):
        return self._calls.copy()

    @hook("after_test")
    async def after_test(self) -> None:
        self.call("after_test")

    @hook("after_test")
    async def after_test2(self) -> None:
        self.call("after_test_2")

    @hook("before_test")
    async def before_test(self) -> None:
        self.call("before_test")

    @hook("before_test")
    async def before_test2(self) -> None:
        self.call("before_test_2")


class DerivedHooked(Hooked):
    @hook("before_test")
    async def derived_before_test(self) -> None:
        super().call("derived_before_test")


class Hooked2(HookObject):
    @hook("before_test")
    async def another_class(self) -> NoReturn:
        raise Exception


class TestHookable:
    @pytest.mark.asyncio()
    async def test_trigger_hook(self) -> None:
        subject = Hooked()
        process = mock.MagicMock()
        async with subject.trigger_hooks("test"):
            process()

        assert subject.calls == [
            "before_test",
            "before_test_2",
            "after_test",
            "after_test_2",
        ]
        process.assert_called_once()

    @pytest.mark.asyncio()
    async def test_trigger_hook_raise(self) -> NoReturn:
        subject = Hooked()
        with pytest.raises(Exception):  # noqa: B017, PT011
            async with subject.trigger_hooks("test"):
                raise Exception

        assert subject.calls == ["before_test", "before_test_2"]

    @pytest.mark.asyncio()
    async def test_trigger_derived_hook(self) -> None:
        subject = DerivedHooked()

        process = mock.MagicMock()
        async with subject.trigger_hooks("test"):
            process()

        assert subject.calls == [
            "before_test",
            "before_test_2",
            "derived_before_test",
            "after_test",
            "after_test_2",
        ]
        process.assert_called_once()



"""Defines `fail_stale_jobs`."""

from __future__ import annotations

import typing as t

import structlog

from .finder import JobFinder

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

logger = structlog.stdlib.get_logger(__name__)


def fail_stale_jobs(session: Session, state_id: str | None = None) -> None:
    """Mark stale jobs as failed.

    Args:
        session: An ORM DB session.
        state_id: If provided, only jobs with this state ID will be failed if stale.
    """
    finder = JobFinder.all_stale if state_id is None else JobFinder(state_id).stale
    for job in finder(session):
        if not job.fail_stale():
            continue

        job.save(session)

        # No need to mention state ID if they're all going to be the same.
        with_state_id = "" if state_id else f" with state ID '{job.job_name}'"

        error = job.payload["error"]
        logger.info(
            f"Marked stale run{with_state_id} that started at "  # noqa: G004
            f"{job.started_at} as failed: {error}",
        )



from __future__ import annotations

import os
import platform
import subprocess
import sys
import typing as t
from pathlib import Path

import mock
import pytest

import meltano
from meltano.core.meltano_invoker import MELTANO_COMMAND, MeltanoInvoker
from meltano.core.tracking.contexts import environment_context

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class TestMeltanoInvoker:
    @pytest.fixture()
    def subject(self, project):
        return MeltanoInvoker(project)

    def test_invoke(self, subject: MeltanoInvoker) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        process = subject.invoke(["--version"], stdout=subprocess.PIPE)
        assert process.returncode == 0
        assert meltano.__version__ in str(process.stdout)

    def test_env(
        self,
        subject: MeltanoInvoker,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        # Process env vars are injected
        monkeypatch.setenv("ENV_VAR_KEY", "ENV_VAR_VALUE_1")
        assert subject._executable_env()["ENV_VAR_KEY"] == "ENV_VAR_VALUE_1"

        # Provided env vars overrides process env vars
        env = subject._executable_env({"ENV_VAR_KEY": "ENV_VAR_VALUE_2"})
        assert env["ENV_VAR_KEY"] == "ENV_VAR_VALUE_2"

        # Environment context UUID from parent process is injected
        assert (
            env["MELTANO_PARENT_CONTEXT_UUID"]
            == environment_context.data["context_uuid"]
        )

    def test_invoke_executable(self, subject: MeltanoInvoker, project: Project) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        process_mock = mock.Mock(returncode=0)
        with mock.patch("subprocess.run", return_value=process_mock) as run_mock:
            subject.invoke(["--version"])

            # Preferally, we use the symlink created by Project.activate
            symlink_path = project.run_dir().joinpath("bin")
            assert run_mock.call_args[0][0][0] == str(symlink_path)

            # If a different command is used...
            subject.invoke(["--version"], command="pip")

            # ...the symlink is not relevant and we find the executable next
            # to the `python` executable
            pip_path = Path(os.path.dirname(sys.executable), "pip")  # noqa: PTH120
            assert run_mock.call_args[0][0][0] == str(pip_path)

            # If the `meltano` symlink does not exist...
            symlink_path.unlink()
            subject.invoke(["--version"])

            # we find the executable next to the `python` executable
            meltano_path = Path(os.path.dirname(sys.executable), MELTANO_COMMAND)  # noqa: PTH120
            assert run_mock.call_args[0][0][0] == str(meltano_path)

            # If the executable doesn't exist in either place...
            subject.invoke(["--version"], command="nonexistent")

            # ...we expect it to be in the PATH
            assert run_mock.call_args[0][0][0] == "nonexistent"



"""Create dedicated state table.

Revision ID: 6828cc5b1a4f
Revises: 5b43800443d1
Create Date: 2022-09-26 12:47:53.512069

"""

from __future__ import annotations

import uuid
from datetime import datetime, timedelta
from enum import Enum

import sqlalchemy
import sqlalchemy as sa
from alembic import op
from sqlalchemy import Column, MetaData, func, select, types
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm.session import Session

from meltano.core.sqlalchemy import GUID, IntFlag, JSONEncodedDict
from meltano.core.utils import merge
from meltano.migrations.utils.dialect_typing import (
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "6828cc5b1a4f"
down_revision = "f4c225a9492f"
branch_labels = None
depends_on = None

SystemMetadata = MetaData()
SystemModel = declarative_base(metadata=SystemMetadata)


# Copied from core/job_state.py
class JobState(SystemModel):
    """Model class that represents the current state of a given job.

    Modified during `meltano elt` or `meltano run` runs whenever a
    STATE message is emitted by a Singer target. Also written and read
    by `meltano state` CLI invocations. Only holds the _current_ state
    for a given job_name. Full job run history is held by the Job model.
    """

    __tablename__ = "state"
    state_id = Column(types.String, unique=True, primary_key=True, nullable=False)

    updated_at = Column(
        types.TIMESTAMP, server_default=func.now(), onupdate=func.current_timestamp()
    )

    partial_state = Column(MutableDict.as_mutable(JSONEncodedDict))
    completed_state = Column(MutableDict.as_mutable(JSONEncodedDict))

    @classmethod
    def from_job_history(cls, session: Session, state_id: str):  # noqa: ANN206, D417
        """Build JobState from job run history.

        Args:
            session: the session to use in finding job history
            job_name: the name of the job to build JobState for

        Returns:
            JobState built from job run history
        """
        completed_state = {}
        partial_state = {}
        incomplete_since = None
        finder = JobFinder(state_id)

        # Get the state for the most recent completed job.
        # Do not consider dummy jobs create via add_state.
        state_job = finder.latest_with_payload(session, flags=Payload.STATE)
        if state_job:
            incomplete_since = state_job.ended_at
            if "singer_state" in state_job.payload:
                merge(state_job.payload, partial_state)

        # If there have been any incomplete jobs since the most recent completed jobs,
        # merge the state emitted by those jobs into the state for the most recent
        # completed job. If there are no completed jobs, get the full history of
        # incomplete jobs and use the most recent state emitted per stream
        incomplete_state_jobs = finder.with_payload(
            session, flags=Payload.INCOMPLETE_STATE, since=incomplete_since
        )
        for incomplete_state_job in incomplete_state_jobs:
            if "singer_state" in incomplete_state_job.payload:
                partial_state = merge(incomplete_state_job.payload, partial_state)
        return cls(
            state_id=state_id,
            partial_state=partial_state,
            completed_state=completed_state,
        )


# Copied from core/job/finder.py
class JobFinder:
    """Query builder for the `Job` model for a certain `elt_uri`."""

    def __init__(self, state_id: str):
        """Initialize the JobFinder.

        Args:
            state_id: the state_id to build queries for.
        """
        self.state_id = state_id

    def latest(self, session):  # noqa: ANN001, ANN201
        """Get the latest state for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest state for this instance's state ID
        """
        return (
            session.query(Job)
            .filter(Job.job_name == self.state_id)
            .order_by(Job.started_at.desc())
            .first()
        )

    def successful(self, session):  # noqa: ANN001, ANN201
        """Get all successful jobs for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            All successful jobs for this instance's state ID
        """
        return session.query(Job).filter(
            (Job.job_name == self.state_id)
            & (Job.state == State.SUCCESS)
            & Job.ended_at.isnot(None)
        )

    def running(self, session):  # noqa: ANN001, ANN201
        """Find states in the running state.

        Args:
            session: the session to use in querying the db

        Returns:
            All runnings states for state_id.
        """
        return session.query(Job).filter(
            (Job.job_name == self.state_id) & (Job.state == State.RUNNING)
        )

    def latest_success(self, session):  # noqa: ANN001, ANN201
        """Get the latest successful state for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest successful state for this instance's state ID
        """
        return self.successful(session).order_by(Job.ended_at.desc()).first()

    def latest_running(self, session):  # noqa: ANN001, ANN201
        """Find the most recent state in the running state, if any.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest running state for this instance's state ID
        """
        return self.running(session).order_by(Job.started_at.desc()).first()

    def with_payload(self, session, flags=0, since=None, state=None):  # noqa: ANN001, ANN201
        """Get all states for this instance's state ID matching the given args.

        Args:
            session: the session to use in querying the db
            flags: only return states with these flags
            since: only return states which ended after this time
            state: only include states with state matching this state

        Returns:
            All states matching these args.
        """
        query = (
            session.query(Job)
            .filter(
                (Job.job_name == self.state_id)
                & (Job.payload_flags != 0)
                & (Job.payload_flags.op("&")(flags) == flags)
                & Job.ended_at.isnot(None)
            )
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)
        if state:
            query = query.filter(Job.state == state)
        return query

    def latest_with_payload(self, session, **kwargs):  # noqa: ANN001, ANN003, ANN201
        """Return the latest state matching the given kwargs.

        Args:
            session: the session to use to query the db
            kwargs: keyword args to pass to with_payload

        Returns:
            The most recent state returned by with_payload(kwargs)
        """
        return (
            self.with_payload(session, **kwargs)
            .order_by(None)  # Reset ascending order
            .order_by(Job.ended_at.desc())
            .first()
        )

    @classmethod
    def all_stale(cls, session):  # noqa: ANN001, ANN206
        """Return all stale states.

        Args:
            session: the session to use to query the db

        Returns:
            All stale states with any state ID
        """
        now = datetime.utcnow()  # noqa: DTZ003
        last_valid_heartbeat_at = now - timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        last_valid_started_at = now - timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return session.query(Job).filter(
            (Job.state == State.RUNNING)
            & (
                (
                    Job.last_heartbeat_at.isnot(None)
                    & (Job.last_heartbeat_at < last_valid_heartbeat_at)
                )
                | (
                    Job.last_heartbeat_at.is_(None)
                    & (Job.started_at < last_valid_started_at)
                )
            )
        )

    def stale(self, session):  # noqa: ANN001, ANN201
        """Return stale states with the instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            All stale states with instance's state ID
        """
        return self.all_stale(session).filter(Job.job_name == self.state_id)

    def get_all(self, session: object, since=None):  # noqa: ANN001, ANN201
        """Return all state with the instance's state ID.

        Args:
            session: the session to use in querying the db
            since: only return state which ended after this datetime

        Returns:
            All state with instance's state ID which ended after 'since'
        """
        query = (
            session.query(Job)
            .filter(Job.job_name == self.state_id)
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)
        return query


# Copied from core/job/job.py
HEARTBEATLESS_JOB_VALID_HOURS = 24
HEARTBEAT_VALID_MINUTES = 5


class State(Enum):
    """Represents status of a Job."""

    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())
    STATE_EDIT = (5, ())


class Payload(IntFlag):
    """Flag indicating whether a Job has state in its payload field."""

    STATE = 1
    INCOMPLETE_STATE = 2


class Job(SystemModel):
    """Model class that represents a `meltano elt` run in the system database.

    Includes State.STATE_EDIT rows which represent CLI invocations of the
    `meltano state` command which wrote state to the db. Queries that are
    meant to return only actual job runs should filter out records with
    state == State.STATE_EDIT.
    """

    __tablename__ = "runs"

    id = Column(types.Integer, primary_key=True)
    job_name = Column(types.String(1024))
    run_id = Column(GUID, nullable=False, default=uuid.uuid4)
    _state = Column(name="state", type_=types.String(10))
    started_at = Column(types.DateTime)
    last_heartbeat_at = Column(types.DateTime)
    ended_at = Column(types.DateTime)
    payload = Column(MutableDict.as_mutable(JSONEncodedDict))
    payload_flags = Column(IntFlag, default=0)
    trigger = Column(types.String, default="")


def upgrade() -> None:
    # Create state table
    dialect_name = get_dialect_name()
    max_string_length = max_string_length_for_dialect(dialect_name)
    conn = op.get_bind()
    inspector = sqlalchemy.inspect(conn)
    if "state" in inspector.get_table_names():
        op.drop_table("state")
    op.create_table(
        "state",
        sa.Column("state_id", sa.String(900), nullable=False),
        sa.Column(
            "partial_state",
            MutableDict.as_mutable(JSONEncodedDict(max_string_length)),
        ),
        sa.Column(
            "completed_state",
            MutableDict.as_mutable(JSONEncodedDict(max_string_length)),
        ),
        sa.Column(
            "updated_at",
            sa.types.TIMESTAMP if dialect_name == "postgresql" else sa.types.DATETIME,
            onupdate=datetime.now,
        ),
        sa.PrimaryKeyConstraint("state_id"),
        sa.UniqueConstraint("state_id"),
    )
    session = Session(bind=conn)
    for state_id in session.execute(select(Job.job_name).distinct()).scalars():
        session.add(JobState.from_job_history(session, state_id))
    session.commit()


def downgrade() -> None:
    # Remove job_state table
    # Job run history is still maintained, so no need to copy
    op.drop_table("state")



"""Plugin Add CLI."""

from __future__ import annotations

import typing as t
from pathlib import Path
from urllib.parse import urlparse

import click
import requests

from meltano.cli.params import InstallPlugins, get_install_options, pass_project
from meltano.cli.utils import (
    CliError,
    PartialInstrumentedCmd,
    add_plugin,
    add_required_plugins,
    check_dependencies_met,
)
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.project_add_service import ProjectAddService
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext
from meltano.core.utils import run_async
from meltano.core.yaml import yaml

if t.TYPE_CHECKING:
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project
    from meltano.core.tracking import Tracker

install, no_install = get_install_options(include_only_install=False)


def _load_yaml_from_ref(_ctx, _param, value: str | None) -> dict | None:
    if not value:
        return None

    try:
        url = urlparse(value)
        if url.scheme and url.netloc:
            response = requests.get(value, timeout=10)
            response.raise_for_status()
            content = response.text
        else:
            content = Path(value).read_text()

    except (ValueError, FileNotFoundError, IsADirectoryError) as e:
        raise click.BadParameter(e) from e

    return yaml.load(content) or {}


@click.command(
    cls=PartialInstrumentedCmd,
    short_help="Add a plugin to your project.",
)
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_name", nargs=-1, required=True)
@click.option(
    "--inherit-from",
    help=(
        "Add a plugin inheriting from an existing plugin in the project"
        " or a discoverable plugin identified, by name."
    ),
)
@click.option(
    "--variant",
    help="Add a specific (non-default) variant of the identified discoverable plugin.",
)
@click.option(
    "--as",
    "as_name",
    help=(
        "Shorthand for '--inherit-from', that can be used to add a "
        "discoverable plugin to your project with a different name. "
        "Usage:\b\n\nadd <type> <inherit-from> --as <name>"
    ),
)
@click.option(
    "--from-ref",
    "plugin_yaml",
    callback=_load_yaml_from_ref,
    help="Reference a plugin definition to add from.",
)
@click.option(
    "--python",
    help=(
        "The Python version to use for the plugin. Only applies to base plugins which "
        "have Python virtual environments, rather than inherited plugins which use the "
        "virtual environment of their base plugin, or plugins that run in a container."
    ),
)
@click.option(
    "--custom",
    is_flag=True,
    help=(
        "Add a custom plugin. The command will prompt you for the package's "
        "base plugin description metadata."
    ),
)
@click.option(
    "--update",
    is_flag=True,
    help="Update an existing plugin.",
)
@install
@no_install
@click.option(
    "--force-install",
    is_flag=True,
    help="Ignore the required Python version declared by the plugins.",
)
@pass_project()
@click.pass_context
@run_async
async def add(
    ctx,  # noqa: ANN001
    project: Project,
    plugin_type: str,
    plugin_name: str,
    install_plugins: InstallPlugins,
    inherit_from: str | None = None,
    variant: str | None = None,
    as_name: str | None = None,
    plugin_yaml: dict | None = None,
    python: str | None = None,
    **flags,  # noqa: ANN003
) -> None:
    """Add a plugin to your project.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#add
    """  # noqa: D301
    tracker: Tracker = ctx.obj["tracker"]

    plugin_type = PluginType.from_cli_argument(plugin_type)
    plugin_names = plugin_name  # nargs=-1

    if as_name:
        # `add <type> <inherit-from> --as <name>``
        # is equivalent to:
        # `add <type> <name> --inherit-from <inherit-from>``
        inherit_from = plugin_names[0]
        plugin_names = [as_name]

    if flags["custom"] and plugin_type in {
        PluginType.TRANSFORMS,
        PluginType.ORCHESTRATORS,
    }:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(f"--custom is not supported for {plugin_type}")  # noqa: EM102

    plugin_refs = [
        PluginRef(plugin_type=plugin_type, name=name) for name in plugin_names
    ]
    dependencies_met, err = check_dependencies_met(
        plugin_refs=plugin_refs,
        plugins_service=project.plugins,
    )
    if not dependencies_met:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(f"Failed to install plugin(s): {err}")  # noqa: EM102

    add_service = ProjectAddService(project)

    plugins: list[ProjectPlugin] = []
    for plugin in plugin_names:
        try:
            plugins.append(
                add_plugin(
                    plugin_type,
                    plugin,
                    python=python,
                    inherit_from=inherit_from,
                    variant=variant,
                    custom=flags["custom"],
                    update=flags["update"],
                    add_service=add_service,
                    plugin_yaml=plugin_yaml,
                ),
            )
        except Exception:
            # if the plugin is not known to meltano send what information we do have
            tracker.add_contexts(
                PluginsTrackingContext([(plugin, None) for plugin in plugins]),
            )
            tracker.track_command_event(CliEvent.aborted)
            raise

        required_plugins = add_required_plugins(
            plugins,
            add_service=add_service,
        )
    plugins.extend(required_plugins)
    tracker.add_contexts(
        PluginsTrackingContext([(candidate, None) for candidate in plugins]),
    )
    tracker.track_command_event(CliEvent.inflight)

    success = await install_plugins(
        project,
        plugins,
        reason=PluginInstallReason.ADD,
        force=flags.get("force_install", False),
    )

    if success is False:
        tracker.track_command_event(CliEvent.failed)
        raise CliError("Failed to install plugin(s)")  # noqa: EM101

    _print_plugins(plugins)
    tracker.track_command_event(CliEvent.completed)


def _print_plugins(plugins) -> None:  # noqa: ANN001
    printed_empty_line = False
    for plugin in plugins:
        docs_url = plugin.docs or plugin.repo
        if not docs_url:
            continue

        if not printed_empty_line:
            click.echo()
            printed_empty_line = True

        click.echo(
            f"To learn more about {plugin.type.descriptor} '{plugin.name}', "
            f"visit {docs_url}",
        )



from __future__ import annotations

import pytest
from mock import AsyncMock

from meltano.core.block.plugin_command import plugin_command_invoker


class TestInvokerCommand:
    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session")
    async def test_run_passes_command_args_when_required(self, project, dbt) -> None:
        cmd = plugin_command_invoker(
            dbt,
            project,
            command="test",
        )

        start_mock = AsyncMock()
        cmd.start = start_mock

        await cmd._start()
        assert not cmd.start.call_args[0]

        cmd = plugin_command_invoker(
            dbt,
            project,
            command="test",
            command_args=["--foo", "--bar"],
        )

        start_mock = AsyncMock()
        cmd.start = start_mock

        await cmd._start()
        assert len(cmd.start.call_args[0]) == 1
        assert cmd.start.call_args[0][0] == ["--foo", "--bar"]



from __future__ import annotations

import os
import platform
import re
import subprocess
import sys
import typing as t
from asyncio.subprocess import Process
from pathlib import Path

import mock
import pytest

from meltano.core.error import AsyncSubprocessError, MeltanoError
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_install_service import install_pip_plugin
from meltano.core.venv_service import UvVenvService, VenvService, VirtualEnv, find_uv

if t.TYPE_CHECKING:
    from meltano.core.project import Project


def _check_venv_created_with_python(project: Project, python: str | None) -> None:
    with mock.patch("meltano.core.venv_service._resolve_python_path") as venv_mock:
        VenvService(project=project)
        venv_mock.assert_called_once_with(python)


async def _check_venv_created_with_python_for_plugin(
    project: Project,
    plugin: ProjectPlugin,
    python: str | None,
) -> None:
    with mock.patch(
        "meltano.core.venv_service._resolve_python_path",
    ) as venv_mock, mock.patch("meltano.core.venv_service.VenvService.install"):
        await install_pip_plugin(project=project, plugin=plugin)
        venv_mock.assert_called_once_with(python)


class TestVenvService:
    @pytest.fixture()
    def subject(self, project):
        return VenvService(project=project, namespace="namespace", name="name")

    def test_clean_run_files(self, project: Project, subject: VenvService) -> None:
        run_dir = project.run_dir("name")

        file = run_dir / "test.file.txt"
        file.touch()

        assert file.exists()
        assert file.is_file()

        sub_dir = run_dir / "test_dir"
        sub_dir.mkdir()

        assert sub_dir.exists()
        assert sub_dir.is_dir()

        subject.clean_run_files()

        assert not file.exists()
        assert not sub_dir.exists()
        assert (
            run_dir.exists()
        ), "Expected all files in the run dir to be removed, but not the dir itself"

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("project")
    async def test_clean_install(self, subject: VenvService) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        await subject.install(["example"], clean=True)
        venv_dir = subject.project.venvs_dir("namespace", "name")

        # ensure the venv is created
        assert venv_dir.exists()

        # ensure that the binary is python3
        assert (venv_dir / "bin/python").samefile(venv_dir / "bin/python3")

        # ensure that the package is installed
        run = subprocess.run(
            [venv_dir.joinpath("bin/python"), "-m", "pip", "list"],
            check=True,
            capture_output=True,
        )
        assert re.search(r"example\s+0\.1\.0", str(run.stdout))

        # ensure that pip is the latest version
        run = subprocess.run(
            [venv_dir.joinpath("bin/python"), "-m", "pip", "list", "--outdated"],
            check=True,
            capture_output=True,
        )
        for line in str(run.stdout).splitlines():
            assert not line.startswith("pip ")

        assert subject.exec_path("some_exe").parts[-6:] == (
            ".meltano",
            "namespace",
            "name",
            "venv",
            "bin",
            "some_exe",
        )

        # ensure a fingerprint file was created
        fingerprint = (venv_dir / ".meltano_plugin_fingerprint").read_text()
        assert (
            fingerprint
            # sha256 of "example"
            == "50d858e0985ecc7f60418aaf0cc5ab587f42c2570a884095a9e8ccacd0f6545c"
        )

        # ensure that log file was created and is not empty
        assert subject.pip_log_path.exists()
        assert subject.pip_log_path.is_file()
        assert subject.pip_log_path.stat().st_size > 0

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("project")
    async def test_install(self, subject: VenvService) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        # Make sure the venv exists already
        await subject.install(["example"], clean=True)
        venv_dir = subject.project.venvs_dir("namespace", "name")

        # remove the package, then check that after a regular install the package exists
        await subject.uninstall_package("example")
        await subject.install(["example"])

        # ensure that the package is installed
        run = subprocess.run(
            [venv_dir.joinpath("bin/python"), "-m", "pip", "list"],
            check=True,
            capture_output=True,
        )
        assert re.search(r"example\s+0\.1\.0", str(run.stdout))

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("project")
    async def test_requires_clean_install(self, subject: VenvService) -> None:
        # Make sure the venv exists already
        await subject.install(["example"], clean=True)

        if platform.system() != "Windows":
            python_path = subject.exec_path("python")
            original_link_target = os.readlink(python_path)
            try:
                # Simulate the deletion of the underlying Python executable by
                # making its symlink point to a file that does not exist
                python_path.unlink()
                python_path.symlink_to("./fake/path/to/python/executable")
                assert subject.requires_clean_install(["example"])
            finally:
                # Restore the symlink to the actual Python executable
                python_path.unlink()
                python_path.symlink_to(original_link_target)

        assert not subject.requires_clean_install(["example"])
        assert subject.requires_clean_install(["example==0.1.0"])
        assert subject.requires_clean_install(["example", "another-package"])

    def test_top_level_python_setting(self, project: Project) -> None:
        project.settings.set("python", "test-python-executable-project-level")
        _check_venv_created_with_python(project, "test-python-executable-project-level")
        project.settings.unset("python")
        _check_venv_created_with_python(project, None)

    async def test_plugin_python_setting(self, project: Project) -> None:
        plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            name="tap-mock",
            python="test-python-executable-plugin-level",
        )

        await _check_venv_created_with_python_for_plugin(
            project,
            plugin,
            "test-python-executable-plugin-level",
        )

        # Setting the project-level `python` setting should have no effect at first
        # because the plugin-level setting takes precedence.
        project.settings.set("python", "test-python-executable-project-level")

        await _check_venv_created_with_python_for_plugin(
            project,
            plugin,
            "test-python-executable-plugin-level",
        )

        # The project-level setting should have an effect after the plugin-level
        # setting is unset
        plugin = ProjectPlugin(PluginType.EXTRACTORS, name="tap-mock")

        await _check_venv_created_with_python_for_plugin(
            project,
            plugin,
            "test-python-executable-project-level",
        )

        project.settings.unset("python")

        # After both the project-level and plugin-level are unset, it should be None
        await _check_venv_created_with_python_for_plugin(project, plugin, None)


class TestVirtualEnv:
    @pytest.mark.parametrize(
        ("system", "lib_dir"),
        (
            ("Linux", "lib"),
            ("Darwin", "lib"),
            ("Windows", "Lib"),
        ),
    )
    def test_cross_platform(self, system: str, lib_dir: str, project: Project) -> None:
        with mock.patch("platform.system", return_value=system):
            subject = VirtualEnv(project.venvs_dir("pytest", "pytest"))
            assert subject.lib_dir == subject.root / lib_dir

    def test_unknown_platform(self, project: Project) -> None:
        with mock.patch("platform.system", return_value="commodore64"), pytest.raises(
            MeltanoError,
            match="(?i)Platform 'commodore64'.*?not supported.",
        ):
            VirtualEnv(project.venvs_dir("pytest", "pytest"))

    def test_different_python_versions(self, project: Project) -> None:
        root = project.venvs_dir("pytest", "pytest")

        assert (
            VirtualEnv(root, python=None).python_path
            == VirtualEnv(root).python_path
            == VirtualEnv(root, python=sys.executable).python_path
            == sys.executable
        )

        with mock.patch(
            "shutil.which",
            return_value="/usr/bin/test-python-executable",
        ), mock.patch("os.access", return_value=True):
            assert (
                VirtualEnv(root, python="test-python-executable").python_path
                == "/usr/bin/test-python-executable"
            )

        with mock.patch("os.path.exists", return_value=True), mock.patch(
            "os.access",
            return_value=True,
        ):
            path_str = "/usr/bin/test-python-executable"
            venv = VirtualEnv(root, python=path_str)
            assert venv.python_path == path_str

            venv = VirtualEnv(root, python=Path(path_str))
            assert venv.python_path == str(Path(path_str).resolve())

        with mock.patch(
            "shutil.which",
            return_value="/usr/bin/test-python-executable",
        ), mock.patch(
            "os.access",
            return_value=False,
        ), pytest.raises(
            MeltanoError,
            match="'/usr/bin/test-python-executable' is not executable",
        ):
            VirtualEnv(root, python="test-python-executable")

        with pytest.raises(
            MeltanoError,
            match="Python executable 'test-python-executable' was not found",
        ):
            VirtualEnv(root, python="test-python-executable")

        with pytest.raises(
            MeltanoError,
            match="not the number 3.11",
        ):
            VirtualEnv(root, python=3.11)


class TestUvVenvService:
    @pytest.fixture()
    def subject(self, project):
        find_uv.cache_clear()
        return UvVenvService(project=project, namespace="namespace", name="name")

    def test_find_uv_builtin(self, monkeypatch: pytest.MonkeyPatch) -> None:
        find_uv.cache_clear()
        monkeypatch.setattr("uv.find_uv_bin", lambda: "/usr/bin/uv")
        assert find_uv() == "/usr/bin/uv"

    def test_find_uv_global(self, monkeypatch: pytest.MonkeyPatch) -> None:
        find_uv.cache_clear()

        def raise_import_error() -> t.NoReturn:
            raise ImportError

        monkeypatch.setattr("uv.find_uv_bin", raise_import_error)
        monkeypatch.setattr("shutil.which", lambda _: "/usr/bin/uv")

        assert find_uv() == "/usr/bin/uv"

    def test_find_uv_not_found(self, monkeypatch: pytest.MonkeyPatch) -> None:
        find_uv.cache_clear()

        def raise_import_error() -> t.NoReturn:
            raise ImportError

        monkeypatch.setattr("uv.find_uv_bin", raise_import_error)
        monkeypatch.setattr("shutil.which", lambda _: None)

        with pytest.raises(MeltanoError, match="Could not find the 'uv' executable"):
            find_uv()

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("project")
    async def test_install(self, subject: UvVenvService) -> None:
        # Make sure the venv exists already
        await subject.install(["cowsay"], clean=True)

        # remove the package, then check that after a regular install the package exists
        await subject.uninstall_package("cowsay")
        await subject.install(["cowsay"])

        run = subprocess.run(
            [
                subject.uv,
                "pip",
                "list",
                "--python",
                str(subject.exec_path("python")),
            ],
            check=True,
            capture_output=True,
        )
        assert "cowsay" in str(run.stdout)

    async def test_handle_installation_error(self, subject: UvVenvService) -> None:
        process = mock.Mock(spec=Process)
        process.stderr = "Some error"

        with mock.patch(
            "meltano.core.venv_service.UvVenvService.install_pip_args",
            side_effect=AsyncSubprocessError("Something went wrong", process),
        ), pytest.raises(
            AsyncSubprocessError,
            match="Failed to install plugin 'name'",
        ):
            await subject.install(["cowsay"])



from __future__ import annotations

import datetime
import platform
import tempfile
from pathlib import Path
from textwrap import dedent

import pytest

from fixtures.utils import cd
from meltano.core.utils import deep_merge


@pytest.fixture()
def cd_temp_subdir():
    original_dir = Path.cwd()
    with tempfile.TemporaryDirectory(dir=original_dir) as name, cd(
        Path(name).resolve(),
    ) as new_dir:
        yield new_dir


@pytest.fixture()
def cd_temp_dir():
    with tempfile.TemporaryDirectory() as name, cd(Path(name).resolve()) as new_dir:
        yield new_dir


@pytest.mark.order(0)
@pytest.mark.parametrize(
    ("parent", "children", "expected"),
    (
        ({"a": 1}, [{"a": 1}], {"a": 1}),
        ({"a": 1}, [{"a": 2}], {"a": 2}),
        ({"a": 1}, [{"a": 2, "b": 2}], {"a": 2, "b": 2}),
        ({"a": [1, 2, 3]}, [{"a": [3, 4, 5]}], {"a": [1, 2, 3, 3, 4, 5]}),
        ({"a": "A", "b": "B"}, [{"a": "Z"}], {"a": "Z", "b": "B"}),
    ),
)
def test_deep_merge(parent, children, expected) -> None:
    assert deep_merge(parent, *children) == expected


class TestProjectFiles:
    @pytest.mark.order(1)
    def test_resolve_subfiles(self, project_files) -> None:
        assert project_files._meltano_file_path == (project_files.root / "meltano.yml")
        assert project_files.meltano == {
            "version": 1,
            "default_environment": "test-meltano-environment",
            "database_uri": "sqlite:///.meltano/meltano.db",
            "include_paths": [
                "./subconfig_[0-9].yml",
                "./*/subconfig_[0-9].yml",
                "./*/**/subconfig_[0-9].yml",
            ],
            "plugins": {
                "extractors": [
                    {
                        "name": "tap-meltano-yml",
                        "settings": [
                            {
                                "name": "token",
                                "description": "Token for the API. This is a secret.",
                                "sensitive": True,
                            },
                        ],
                    },
                ],
                "mappers": [
                    {
                        "name": "map-meltano-yml",
                        "mappings": [{"name": "transform-meltano-yml"}],
                    },
                ],
                "loaders": [{"name": "target-meltano-yml"}],
            },
            "schedules": [
                {
                    "name": "test-meltano-yml",
                    "extractor": "tap-meltano-yml",
                    "loader": "target-meltano-yml",
                    "transform": "skip",
                    "interval": "@daily",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        5,
                        0,
                        0,
                        tzinfo=datetime.timezone.utc,
                    ),
                },
            ],
            "environments": [
                {"name": "test-meltano-environment", "env": {"TEST": "TEST-MELTANO"}},
            ],
            "jobs": [
                {
                    "name": "my-job",
                    "tasks": [
                        "tap-meltano-yml map-meltano-yml target-meltano-yml",
                    ],
                },
            ],
        }
        assert project_files.include_paths == [
            (project_files.root / "subconfig_2.yml"),
            (project_files.root / "subfolder" / "subconfig_1.yml"),
        ]

    @pytest.mark.order(2)
    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="Test fails if even attempted to be run, xfail can't save us here.",
    )
    def test_resolve_from_subdir(self, project_files, cd_temp_subdir) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        assert Path.cwd() == cd_temp_subdir
        assert cd_temp_subdir.parent == project_files.root
        assert project_files.include_paths == [
            (project_files.root / "subconfig_2.yml"),
            (project_files.root / "subfolder" / "subconfig_1.yml"),
        ]

    @pytest.mark.order(3)
    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="Test fails if even attempted to be run, xfail can't save us here.",
    )
    def test_resolve_from_any_dir(self, project_files, cd_temp_dir) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        assert Path.cwd() == cd_temp_dir
        assert project_files.include_paths == [
            (project_files.root / "subconfig_2.yml"),
            (project_files.root / "subfolder" / "subconfig_1.yml"),
        ]

    @pytest.mark.order(5)
    def test_load(self, project_files) -> None:
        expected_result = {
            "version": 1,
            "default_environment": "test-meltano-environment",
            "database_uri": "sqlite:///.meltano/meltano.db",
            "include_paths": [
                "./subconfig_[0-9].yml",
                "./*/subconfig_[0-9].yml",
                "./*/**/subconfig_[0-9].yml",
            ],
            "plugins": {
                "extractors": [
                    {
                        "name": "tap-meltano-yml",
                        "settings": [
                            {
                                "name": "token",
                                "description": "Token for the API. This is a secret.",
                                "sensitive": True,
                            },
                        ],
                    },
                    {"name": "tap-subconfig-2-yml"},
                    {"name": "tap-subconfig-1-yml"},
                ],
                "mappers": [
                    {
                        "name": "map-meltano-yml",
                        "mappings": [{"name": "transform-meltano-yml"}],
                    },
                ],
                "loaders": [
                    {"name": "target-meltano-yml"},
                    {"name": "target-subconfig-2-yml"},
                    {"name": "target-subconfig-1-yml"},
                ],
            },
            "schedules": [
                {
                    "name": "test-meltano-yml",
                    "extractor": "tap-meltano-yml",
                    "loader": "target-meltano-yml",
                    "transform": "skip",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        5,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "interval": "@daily",
                },
                {
                    "name": "test-subconfig-2-yml",
                    "extractor": "tap-subconfig-2-yml",
                    "loader": "target-subconfig-2-yml",
                    "transform": "skip",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        4,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "interval": "@daily",
                },
                {
                    "name": "test-subconfig-1-yml",
                    "extractor": "tap-subconfig-1-yml",
                    "loader": "target-subconfig-1-yml",
                    "transform": "skip",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        6,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "interval": "@daily",
                },
            ],
            "environments": [
                {"name": "test-meltano-environment", "env": {"TEST": "TEST-MELTANO"}},
                {
                    "name": "test-subconfig-2-yml",
                    "env": {"TEST": "TEST-SUBCONFIG-2-YML"},
                },
                {
                    "name": "test-subconfig-1-yml",
                    "env": {"TEST": "TEST-SUBCONFIG-1-YML"},
                },
            ],
            "jobs": [
                {
                    "name": "my-job",
                    "tasks": [
                        "tap-meltano-yml map-meltano-yml target-meltano-yml",
                    ],
                },
            ],
        }
        read_result = project_files.load()
        assert read_result == expected_result

    @pytest.mark.order(6)
    def test_update(self, project_files) -> None:
        meltano_config = project_files.load()
        meltano_config["version"] = 2
        meltano_config["plugins"]["extractors"][1]["name"] = (
            "modified-tap-subconfig-2-yml"
        )
        meltano_config["plugins"]["loaders"][2]["name"] = (
            "modified-target-subconfig-1-yml"
        )
        meltano_config["schedules"][0]["name"] = "modified-test-meltano-yml"

        project_files.update(meltano_config)
        expected_result = {
            "default_environment": "test-meltano-environment",
            "database_uri": "sqlite:///.meltano/meltano.db",
            "include_paths": [
                "./subconfig_[0-9].yml",
                "./*/subconfig_[0-9].yml",
                "./*/**/subconfig_[0-9].yml",
            ],
            "plugins": {
                "extractors": [
                    {
                        "name": "tap-meltano-yml",
                        "settings": [
                            {
                                "name": "token",
                                "description": "Token for the API. This is a secret.",
                                "sensitive": True,
                            },
                        ],
                    },
                    {"name": "modified-tap-subconfig-2-yml"},
                    {"name": "tap-subconfig-1-yml"},
                ],
                "mappers": [
                    {
                        "name": "map-meltano-yml",
                        "mappings": [{"name": "transform-meltano-yml"}],
                    },
                ],
                "loaders": [
                    {"name": "target-meltano-yml"},
                    {"name": "modified-target-subconfig-1-yml"},
                    {"name": "target-subconfig-2-yml"},
                ],
            },
            "schedules": [
                {
                    "extractor": "tap-meltano-yml",
                    "interval": "@daily",
                    "loader": "target-meltano-yml",
                    "name": "modified-test-meltano-yml",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        5,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "transform": "skip",
                },
                {
                    "extractor": "tap-subconfig-2-yml",
                    "interval": "@daily",
                    "loader": "target-subconfig-2-yml",
                    "name": "test-subconfig-2-yml",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        4,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "transform": "skip",
                },
                {
                    "extractor": "tap-subconfig-1-yml",
                    "interval": "@daily",
                    "loader": "target-subconfig-1-yml",
                    "name": "test-subconfig-1-yml",
                    "start_date": datetime.datetime(
                        2020,
                        8,
                        6,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "transform": "skip",
                },
            ],
            "environments": [
                {"name": "test-meltano-environment", "env": {"TEST": "TEST-MELTANO"}},
                {
                    "name": "test-subconfig-2-yml",
                    "env": {"TEST": "TEST-SUBCONFIG-2-YML"},
                },
                {
                    "name": "test-subconfig-1-yml",
                    "env": {"TEST": "TEST-SUBCONFIG-1-YML"},
                },
            ],
            "jobs": [
                {
                    "name": "my-job",
                    "tasks": [
                        "tap-meltano-yml map-meltano-yml target-meltano-yml",
                    ],
                },
            ],
            "version": 2,
        }
        read_result = project_files.load()
        assert read_result == expected_result

    @pytest.mark.order(7)
    def test_preserve_format(self, project_files) -> None:
        meltano_config = project_files.load()
        meltano_config["version"] = 3
        meltano_config["schedules"][0]["transform"] = "only"
        meltano_config["schedules"][0].yaml_add_eol_comment(
            "Only update dbt models\n",
            "transform",
        )

        project_files.update(meltano_config)

        contents = project_files._meltano_file_path.read_text()

        expected_contents = """\
            # Top-level comment
            version: 3
            default_environment: test-meltano-environment
            database_uri: sqlite:///.meltano/meltano.db

            include_paths:
            - ./subconfig_[0-9].yml
            # Config files inside a subfolder
            - ./*/subconfig_[0-9].yml
            - ./*/**/subconfig_[0-9].yml

            schedules:
            # My schedules
            - name: modified-test-meltano-yml
              extractor: tap-meltano-yml
              loader: target-meltano-yml
              transform: only  # Only update dbt models
              start_date: 2020-08-05T00:00:00Z
              interval: '@daily' # Run daily

            jobs:  # My jobs
            # An EL job with mapping
            - name: my-job
              tasks:
              - >-
                tap-meltano-yml
                map-meltano-yml
                target-meltano-yml

            environments:
            # My meltano environments
            - name: test-meltano-environment
              env:
                TEST: TEST-MELTANO

            plugins:
              # Project plugins
              extractors:
              - name: tap-meltano-yml # Comment on array element
                settings:
                - name: token
                  description: >-
                    Token for the API.
                    This is a secret.
                  sensitive: true

              - name: modified-tap-subconfig-2-yml

              mappers:
              # My mappers
              - name: map-meltano-yml
                # These are some mappings
                mappings:
                - name: transform-meltano-yml

              loaders:
              - name: target-meltano-yml
              - name: modified-target-subconfig-1-yml

        """

        assert contents == dedent(expected_contents)

        expected_subconfig_2_contents = """\
            plugins:
              # Subconfig 2 Plugins
              loaders:
              - name: target-subconfig-2-yml  # Subconfig 2 Loader

            schedules:
            # Subconfig 2 Schedules
            - name: test-subconfig-2-yml
              extractor: tap-subconfig-2-yml
              loader: target-subconfig-2-yml
              transform: skip
              start_date: 2020-08-04T00:00:00Z
              interval: '@daily' # Run daily

            environments:
            # Subconfig 2 Environments
            - name: test-subconfig-2-yml
              env:
                TEST: TEST-SUBCONFIG-2-YML
        """

        included_path = project_files.root / "subconfig_2.yml"
        assert included_path.read_text() == dedent(expected_subconfig_2_contents)

    @pytest.mark.order(-1)
    def test_remove_all_file_contents(self, project_files) -> None:
        meltano_config = project_files.load()
        meltano_config["plugins"]["extractors"] = [
            entry
            for entry in meltano_config["plugins"]["extractors"]
            if entry["name"] != "tap-subconfig-2-yml"
        ]
        meltano_config["plugins"]["loaders"] = [
            entry
            for entry in meltano_config["plugins"]["loaders"]
            if entry["name"] != "target-subconfig-2-yml"
        ]
        meltano_config["schedules"] = [
            entry
            for entry in meltano_config["schedules"]
            if entry["name"] != "test-subconfig-2-yml"
        ]
        meltano_config["environments"] = [
            entry
            for entry in meltano_config["environments"]
            if entry["name"] != "test-subconfig-2-yml"
        ]
        project_files.update(meltano_config)

        expected_subconfig_2_contents = """\
        plugins: {}
        schedules: []
        jobs: []
        environments: []
        """

        included_path = project_files.root / "subconfig_2.yml"
        assert included_path.read_text() == dedent(expected_subconfig_2_contents)



from __future__ import annotations

import os

import pytest

from meltano.core.plugin import PluginType, Variant
from meltano.core.project_plugins_service import PluginAlreadyAddedException


def assert_extractor_env(extractor, env) -> None:
    assert env["MELTANO_EXTRACTOR_NAME"] == extractor.name
    assert env["MELTANO_EXTRACTOR_NAMESPACE"] == extractor.namespace
    assert env["MELTANO_EXTRACTOR_VARIANT"] == extractor.variant

    assert env["MELTANO_EXTRACT_TEST"] == env["TAP_MOCK_TEST"] == "mock"
    assert env["MELTANO_EXTRACT__SELECT"] == env["TAP_MOCK__SELECT"] == '["*.*"]'


def assert_loader_env(loader, env) -> None:
    assert env["MELTANO_LOADER_NAME"] == loader.name
    assert env["MELTANO_LOADER_NAMESPACE"] == loader.namespace
    assert env["MELTANO_LOADER_VARIANT"] == loader.variant

    assert env["MELTANO_LOAD_HOST"] == os.getenv("TARGET_POSTGRES_HOST", "localhost")

    assert (
        env["MELTANO_LOAD_DEFAULT_TARGET_SCHEMA"]
        == env["MELTANO_EXTRACT__LOAD_SCHEMA"]
        == env["MELTANO_EXTRACTOR_NAMESPACE"]
    )


def assert_transform_env(transform, env) -> None:
    assert env["MELTANO_TRANSFORM_NAME"] == transform.name
    assert env["MELTANO_TRANSFORM_NAMESPACE"] == transform.namespace
    assert env["MELTANO_TRANSFORM_VARIANT"] == Variant.ORIGINAL_NAME


def assert_transformer_env(transformer, env) -> None:
    assert env["MELTANO_TRANSFORMER_NAME"] == transformer.name
    assert env["MELTANO_TRANSFORMER_NAMESPACE"] == transformer.namespace
    assert env["MELTANO_TRANSFORMER_VARIANT"] == "dbt-labs"

    assert (
        env["MELTANO_TRANSFORM_TARGET"]
        == env["DBT_TARGET"]
        == env["MELTANO_LOAD__DIALECT"]
    )
    assert (
        env["MELTANO_TRANSFORM_TARGET_SCHEMA"]
        == env["DBT_TARGET_SCHEMA"]
        == "analytics"
    )
    assert env["MELTANO_EXTRACT__LOAD_SCHEMA"] == env["MELTANO_EXTRACTOR_NAMESPACE"]
    assert env["MELTANO_TRANSFORM_MODELS"] == env["DBT_MODELS"]

    assert env["MELTANO_TRANSFORM__PACKAGE_NAME"] in env["DBT_MODELS"]
    assert env["MELTANO_EXTRACTOR_NAMESPACE"] in env["DBT_MODELS"]
    assert "my_meltano_project" in env["DBT_MODELS"]


class TestELTContext:
    @pytest.fixture()
    def target_postgres(self, project_add_service):
        try:
            return project_add_service.add(PluginType.LOADERS, "target-postgres")
        except PluginAlreadyAddedException as err:
            return err.plugin

    @pytest.fixture()
    def tap_mock_transform(self, project_add_service):
        try:
            return project_add_service.add(PluginType.TRANSFORMS, "tap-mock-transform")
        except PluginAlreadyAddedException as err:
            return err.plugin

    @pytest.fixture()
    def elt_context(
        self,
        elt_context_builder,
        session,
        tap,
        target_postgres,
        tap_mock_transform,
        dbt,  # noqa: ARG002
    ):
        return (
            elt_context_builder.with_session(session)
            .with_extractor(tap.name)
            .with_loader(target_postgres.name)
            .with_transform(tap_mock_transform.name)
            .with_select_filter(["entity", "!other_entity"])
            .context()
        )

    @pytest.mark.asyncio()
    async def test_extractor(self, elt_context, session, tap) -> None:
        extractor = elt_context.extractor
        assert extractor.type == PluginType.EXTRACTORS
        assert extractor.name == tap.name

        invoker = elt_context.extractor_invoker()
        async with invoker.prepared(session):
            env = invoker.env()

        assert_extractor_env(extractor, env)

    @pytest.mark.asyncio()
    async def test_loader(self, elt_context, session, target_postgres) -> None:
        loader = elt_context.loader
        assert loader.type == PluginType.LOADERS
        assert loader.name == target_postgres.name

        invoker = elt_context.loader_invoker()
        async with invoker.prepared(session):
            env = invoker.env()

        assert_extractor_env(elt_context.extractor, env)
        assert_loader_env(loader, env)

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("target_postgres")
    async def test_transformer(
        self,
        elt_context,
        session,
        tap_mock_transform,
        dbt,
    ) -> None:
        transformer = elt_context.transformer
        assert transformer.type == PluginType.TRANSFORMERS
        assert transformer.name == dbt.name

        transform = elt_context.transform
        assert transform.type == PluginType.TRANSFORMS
        assert transform.name == tap_mock_transform.name

        invoker = elt_context.transformer_invoker()
        async with invoker.prepared(session):
            env = invoker.env()

        assert_extractor_env(elt_context.extractor, env)
        assert_loader_env(elt_context.loader, env)

        assert_transform_env(transform, env)
        assert_transformer_env(transformer, env)

    @pytest.mark.asyncio()
    async def test_select_filter(self, elt_context, session) -> None:
        assert elt_context.select_filter

        invoker = elt_context.extractor_invoker()
        await invoker.prepare(session)
        assert (
            invoker.plugin_config_extras["_select_filter"] == elt_context.select_filter
        )



"""Defines `meltano elt` command."""

from __future__ import annotations

import logging
import platform
import typing as t
from contextlib import asynccontextmanager, nullcontext, suppress
from datetime import datetime, timezone

import click
from structlog import stdlib as structlog_stdlib

from meltano.cli.params import (
    InstallPlugins,
    UUIDParamType,
    get_install_options,
    pass_project,
)
from meltano.cli.utils import CliEnvironmentBehavior, CliError, PartialInstrumentedCmd
from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.job import Job, JobFinder
from meltano.core.job.stale_job_failer import fail_stale_jobs
from meltano.core.logging import JobLoggingService, OutputLogger
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.runner import RunnerError
from meltano.core.runner.dbt import DbtRunner
from meltano.core.runner.singer import SingerRunner
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext
from meltano.core.utils import run_async

if t.TYPE_CHECKING:
    import uuid

    import structlog
    from sqlalchemy.orm import Session

    from meltano.core.project import Project
    from meltano.core.tracking import Tracker

DUMPABLES = {
    "catalog": (PluginType.EXTRACTORS, "catalog"),
    "state": (PluginType.EXTRACTORS, "state"),
    "extractor-config": (PluginType.EXTRACTORS, "config"),
    "loader-config": (PluginType.LOADERS, "config"),
}

logger = structlog_stdlib.get_logger(__name__)

install, no_install, only_install = get_install_options(include_only_install=True)


class ELOptions:
    """Namespace for all options shared by the `el` and `elt` commands."""

    extractor = click.argument("extractor")
    loader = click.argument("loader")
    transform = click.option(
        "--transform",
        type=click.Choice(["skip", "only", "run"]),
    )
    dry = click.option("--dry", help="Do not actually run.", is_flag=True)
    full_refresh = click.option(
        "--full-refresh",
        help="Perform a full refresh (ignore state left behind by any previous runs).",
        is_flag=True,
    )
    refresh_catalog = click.option(
        "--refresh-catalog",
        help="Invalidates catalog cache and forces running discovery before this run.",
        is_flag=True,
    )
    select = click.option(
        "--select",
        "-s",
        help="Select only these specific entities for extraction.",
        multiple=True,
        default=[],
    )
    exclude = click.option(
        "--exclude",
        "-e",
        help="Exclude these specific entities from extraction.",
        multiple=True,
        default=[],
    )
    catalog = click.option("--catalog", help="Extractor catalog file.")
    state = click.option("--state", help="Extractor state file.")
    dump = click.option(
        "--dump",
        type=click.Choice(tuple(DUMPABLES)),
        help="Dump content of pipeline-specific generated file.",
    )
    state_id = click.option(
        "--state-id",
        envvar="MELTANO_STATE_ID",
        help="A custom string to identify the job.",
    )
    force = click.option(
        "--force",
        "-f",
        help=(
            "Force a new run even when a pipeline with the same state ID is "
            "already running."
        ),
        is_flag=True,
    )
    merge_state = click.option(
        "--merge-state",
        is_flag=True,
        help="Merges state with that of previous runs.",
    )
    run_id = click.option(
        "--run-id",
        type=UUIDParamType(),
        help="Provide a UUID to use as the run ID.",
    )


@click.command(
    cls=PartialInstrumentedCmd,
    short_help="Run an EL pipeline to Extract and Load data.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@ELOptions.extractor
@ELOptions.loader
@ELOptions.dry
@ELOptions.full_refresh
@ELOptions.refresh_catalog
@ELOptions.select
@ELOptions.exclude
@ELOptions.catalog
@ELOptions.state
@ELOptions.dump
@ELOptions.state_id
@ELOptions.force
@ELOptions.merge_state
@install
@no_install
@only_install
@ELOptions.run_id
@click.pass_context
@pass_project(migrate=True)
@run_async
async def el(  # WPS408
    project: Project,
    ctx: click.Context,
    *,
    extractor: str,
    loader: str,
    dry: bool,
    full_refresh: bool,
    refresh_catalog: bool,
    select: list[str],
    exclude: list[str],
    catalog: str | None,
    state: str | None,
    dump: str | None,
    state_id: str | None,
    force: bool,
    merge_state: bool,
    run_id: uuid.UUID | None,
    install_plugins: InstallPlugins,
) -> None:
    """Run an EL pipeline to Extract and Load data.

    meltano el '<extractor_name>' '<loader_name>'

    extractor_name: extractor to be used in this pipeline.
    loader_name: loader to be used in this pipeline.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#el
    """  # noqa: D301
    await _run_el_command(
        project=project,
        ctx=ctx,
        extractor=extractor,
        loader=loader,
        transform=None,
        dry=dry,
        full_refresh=full_refresh,
        refresh_catalog=refresh_catalog,
        select=select,
        exclude=exclude,
        catalog=catalog,
        state=state,
        dump=dump,
        state_id=state_id,
        force=force,
        merge_state=merge_state,
        install_plugins=install_plugins,
        run_id=run_id,
    )


@click.command(
    cls=PartialInstrumentedCmd,
    short_help="Run an ELT pipeline to Extract, Load, and Transform data.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@ELOptions.extractor
@ELOptions.loader
@ELOptions.transform
@ELOptions.dry
@ELOptions.full_refresh
@ELOptions.refresh_catalog
@ELOptions.select
@ELOptions.exclude
@ELOptions.catalog
@ELOptions.state
@ELOptions.dump
@ELOptions.state_id
@ELOptions.force
@ELOptions.merge_state
@install
@no_install
@only_install
@ELOptions.run_id
@click.pass_context
@pass_project(migrate=True)
@run_async
async def elt(  # WPS408
    project: Project,
    ctx: click.Context,
    *,
    extractor: str,
    loader: str,
    transform: str,
    dry: bool,
    full_refresh: bool,
    refresh_catalog: bool,
    select: list[str],
    exclude: list[str],
    catalog: str | None,
    state: str | None,
    dump: str | None,
    state_id: str | None,
    force: bool,
    merge_state: bool,
    install_plugins: InstallPlugins,
    run_id: uuid.UUID | None,
) -> None:
    """Run an ELT pipeline to Extract, Load, and Transform data.

    meltano elt '<extractor_name>' '<loader_name>'

    extractor_name: extractor to be used in this pipeline.
    loader_name: loader to be used in this pipeline.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#elt
    """  # noqa: D301
    logger.warning("The `elt` command is deprecated in favor of `el`")
    await _run_el_command(
        project=project,
        ctx=ctx,
        extractor=extractor,
        loader=loader,
        transform=transform,
        dry=dry,
        full_refresh=full_refresh,
        refresh_catalog=refresh_catalog,
        select=select,
        exclude=exclude,
        catalog=catalog,
        state=state,
        dump=dump,
        state_id=state_id,
        force=force,
        merge_state=merge_state,
        install_plugins=install_plugins,
        run_id=run_id,
    )


async def _run_el_command(
    project: Project,
    ctx: click.Context,
    *,
    extractor: str,
    loader: str,
    transform: str | None,
    dry: bool,
    full_refresh: bool,
    refresh_catalog: bool,
    select: list[str],
    exclude: list[str],
    catalog: str | None,
    state: str | None,
    dump: str | None,
    state_id: str | None,
    force: bool,
    merge_state: bool,
    install_plugins: InstallPlugins,
    run_id: uuid.UUID | None,
) -> None:
    if platform.system() == "Windows":
        raise CliError(
            "ELT command not supported on Windows. Please use the run command "  # noqa: EM101
            "as documented here: "
            "https://docs.meltano.com/reference/command-line-interface#run",
        )

    tracker: Tracker = ctx.obj["tracker"]

    # We no longer set a default choice for transform, so that we can detect
    # explicit usages of the `--transform` option if transform is `None` we
    # still need manually default to skip after firing the tracking event above
    # TODO: Remove the transform option entirely in a future release
    transform = transform or "skip"

    select_filter = [*select, *(f"!{entity}" for entity in exclude)]

    job = Job(
        job_name=state_id
        or (
            f'{datetime.now(timezone.utc).strftime("%Y-%m-%dT%H%M%S")}'
            f"--{extractor}--{loader}"
        ),
        run_id=run_id,
    )
    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        context_builder = _elt_context_builder(
            project,
            job,
            session,
            extractor,
            loader,
            transform,
            dry_run=dry,
            full_refresh=full_refresh,
            refresh_catalog=refresh_catalog,
            select_filter=select_filter,
            catalog=catalog,
            state=state,
            merge_state=merge_state,
            run_id=run_id,
        )

        if dump:
            await dump_file(context_builder, dump)
        else:
            await _run_job(
                tracker,
                project,
                job,
                session,
                context_builder,
                install_plugins,
                force=force,
            )
    except Exception as err:
        tracker.track_command_event(CliEvent.failed)
        raise err
    finally:
        session.close()

    tracker.track_command_event(CliEvent.completed)


def _elt_context_builder(
    project: Project,
    job: Job,
    session: Session,
    extractor: str,
    loader: str,
    transform: str,
    *,
    dry_run: bool = False,
    full_refresh: bool = False,
    refresh_catalog: bool = False,
    select_filter: list[str] | None = None,
    catalog: str | None = None,
    state: str | None = None,
    merge_state: bool = False,
    run_id: uuid.UUID | None = None,
) -> ELTContextBuilder:
    select_filter = select_filter or []
    transform_name = None
    if transform != "skip":
        transform_name = _find_transform_for_extractor(project, extractor)

    return (
        ELTContextBuilder(project)
        .with_session(session)
        .with_job(job)
        .with_extractor(extractor)
        .with_loader(loader)
        .with_transform(transform_name or transform)
        .with_dry_run(dry_run=dry_run)
        .with_only_transform(only_transform=(transform == "only"))
        .with_full_refresh(full_refresh=full_refresh)
        .with_refresh_catalog(refresh_catalog=refresh_catalog)
        .with_select_filter(select_filter)
        .with_catalog(catalog)
        .with_state(state)
        .with_merge_state(merge_state=merge_state)
        .with_run_id(run_id)
    )


async def dump_file(context_builder, dumpable) -> None:  # noqa: ANN001
    """Dump the given dumpable for this pipeline."""
    elt_context = context_builder.context()

    try:
        plugin_type, file_id = DUMPABLES[dumpable]
        invoker = elt_context.invoker_for(plugin_type)

        async with invoker.prepared(elt_context.session):
            content = await invoker.dump(file_id)

        click.echo(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {dumpable} file for this pipeline") from err  # noqa: EM102
    except Exception as err:
        raise CliError(f"Could not dump {dumpable}: {err}") from err  # noqa: EM102


async def _run_job(
    tracker,  # noqa: ANN001
    project,  # noqa: ANN001
    job,  # noqa: ANN001
    session,  # noqa: ANN001
    context_builder,  # noqa: ANN001
    install_plugins: InstallPlugins,
    *,
    force=False,  # noqa: ANN001
) -> None:
    fail_stale_jobs(session, job.job_name)

    if not force and (existing := JobFinder(job.job_name).latest_running(session)):
        raise CliError(
            f"Another '{job.job_name}' pipeline is already running which "  # noqa: EM102
            f"started at {existing.started_at}. To ignore this check use "
            "the '--force' option.",
        )

    async with job.run(session):
        job_logging_service = JobLoggingService(project)
        log_file = job_logging_service.generate_log_name(job.job_name, job.run_id)

        output_logger = OutputLogger(log_file)
        context_builder.set_base_output_logger(output_logger)

        log = logger.bind(name="meltano", run_id=str(job.run_id), state_id=job.job_name)

        await _run_elt(tracker, log, context_builder, output_logger, install_plugins)


@asynccontextmanager
async def _redirect_output(log, output_logger):  # noqa: ANN001, ANN202
    meltano_stdout = output_logger.out(
        "meltano",
        log.bind(stdio="stdout", cmd_type="elt"),
    )
    meltano_stderr = output_logger.out(
        "meltano",
        log.bind(stdio="stderr", cmd_type="elt"),
    )

    with meltano_stdout.redirect_logging(ignore_errors=(CliError,)):
        async with meltano_stdout.redirect_stdout(), meltano_stderr.redirect_stderr():
            try:
                yield
            except CliError as err:
                err.print()
                raise


async def _run_elt(
    tracker: Tracker,
    log: structlog.stdlib.BoundLogger,
    context_builder: ELTContextBuilder,
    output_logger: OutputLogger,
    install_plugins: InstallPlugins,
) -> None:
    elt_context = context_builder.context()
    plugins = [elt_context.extractor, elt_context.loader]

    if elt_context.only_transform:
        plugins.append(elt_context.transformer)

    await install_plugins(
        elt_context.project,
        plugins,
        reason=PluginInstallReason.AUTO,
    )

    async with _redirect_output(log, output_logger):
        try:
            tracker.add_contexts(PluginsTrackingContext.from_elt_context(elt_context))
            tracker.track_command_event(CliEvent.inflight)

            if elt_context.only_transform:
                log.info("Extract & load skipped.")
            else:
                await _run_extract_load(log, elt_context, output_logger)

            if elt_context.transformer:
                log.warning(
                    "The option to run a transformation is deprecated and will be "
                    "removed in a future version.",
                )
                await _run_transform(log, elt_context, output_logger)
            else:
                log.info("Transformation skipped.")
        except RunnerError as err:
            raise CliError(
                f"ELT could not be completed: {err}.\nFor more detailed log "  # noqa: EM102
                "messages re-run the command using 'meltano "
                "--log-level=debug ...' CLI flag.\nNote that you can also "
                f"check the generated log file at '{output_logger.file}'.\n"
                "For more information on debugging and logging: "
                "https://docs.meltano.com/reference/command-line-interface#debugging",
            ) from err


async def _run_extract_load(
    log,  # noqa: ANN001
    elt_context,  # noqa: ANN001
    output_logger,  # noqa: ANN001
    **kwargs,  # noqa: ANN003
) -> None:
    extractor = elt_context.extractor.name
    loader = elt_context.loader.name

    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
    )

    extractor_log = output_logger.out(extractor, stderr_log.bind(cmd_type="extractor"))
    loader_log = output_logger.out(loader, stderr_log.bind(cmd_type="loader"))

    extractor_out_writer_ctxmgr = nullcontext
    loader_out_writer_ctxmgr = nullcontext
    if logger.getEffectiveLevel() == logging.DEBUG:
        stdout_log = logger.bind(
            run_id=str(elt_context.job.run_id),
            state_id=elt_context.job.job_name,
            stdio="stdout",
        )
        extractor_out = output_logger.out(
            f"{extractor} (out)",
            stdout_log.bind(cmd_type="extractor"),
            logging.DEBUG,
        )
        loader_out = output_logger.out(
            f"{loader} (out)",
            stdout_log.bind(cmd_type="loader"),
            logging.DEBUG,
        )

        extractor_out_writer_ctxmgr = extractor_out.line_writer
        loader_out_writer_ctxmgr = loader_out.line_writer

    log.info(
        "Running extract & load...",
    )

    singer_runner = SingerRunner(elt_context)
    try:
        # Once Python 3.9 support has been dropped, consolidate these
        # with-statements by using parentheses.
        with extractor_log.line_writer() as extractor_log_writer:  # noqa: SIM117
            with loader_log.line_writer() as loader_log_writer:
                with extractor_out_writer_ctxmgr() as extractor_out_writer:
                    with loader_out_writer_ctxmgr() as loader_out_writer:
                        await singer_runner.run(
                            **kwargs,
                            extractor_log=extractor_log_writer,
                            loader_log=loader_log_writer,
                            extractor_out=extractor_out_writer,
                            loader_out=loader_out_writer,
                        )
    except RunnerError as err:
        with suppress(KeyError):
            code = err.exitcodes[PluginType.EXTRACTORS]
            message = extractor_log.last_line.rstrip() or "(see above)"
            log.error("Extraction failed", code=code, message=message)
        with suppress(KeyError):
            code = err.exitcodes[PluginType.LOADERS]
            message = loader_log.last_line.rstrip() or "(see above)"
            log.error("Loading failed", code=code, message=message)
        raise

    log.info("Extract & load complete!")


async def _run_transform(log, elt_context, output_logger, **kwargs) -> None:  # noqa: ANN001, ANN003
    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
        cmd_type="transformer",
    )

    transformer_log = output_logger.out(elt_context.transformer.name, stderr_log)

    log.info("Running transformation...")

    dbt_runner = DbtRunner(elt_context)
    try:
        with transformer_log.line_writer() as transformer_log_writer:
            await dbt_runner.run(**kwargs, log=transformer_log_writer)
    except RunnerError as err:
        with suppress(KeyError):
            code = err.exitcodes[PluginType.TRANSFORMERS]
            message = transformer_log.last_line.rstrip() or "(see above)"
            log.error("Transformation failed", code=code, message=message)
        raise

    log.info("Transformation complete!")


def _find_extractor(project: Project, extractor_name: str):  # noqa: ANN202
    try:
        return project.plugins.locked_definition_service.find_definition(
            PluginType.EXTRACTORS,
            extractor_name,
        )
    except PluginNotFoundError:
        return project.hub_service.find_definition(
            PluginType.EXTRACTORS,
            extractor_name,
        )


def _find_transform_for_extractor(  # noqa: ANN202
    project: Project,
    extractor_name: str,
):
    try:
        # Check if there is a default transform for this extractor
        transform_plugin_def = project.plugins.find_plugin_by_namespace(
            PluginType.TRANSFORMS,
            _find_extractor(project, extractor_name).namespace,
        )

        # Check if the transform has been added to the project
        transform_plugin = project.plugins.get_plugin(transform_plugin_def)

        return transform_plugin.name
    except PluginNotFoundError:
        return None



from __future__ import annotations  # noqa: D104

from .base import (
    BasePlugin,
    PluginDefinition,
    PluginRef,
    PluginType,
    StandalonePlugin,
    Variant,
    VariantNotFoundError,
)



"""CLI command `meltano compile`."""

from __future__ import annotations

import json
import typing as t
from pathlib import Path

import click

from meltano.cli.params import pass_project
from meltano.cli.utils import CliError, InstrumentedCmd
from meltano.core.environment import Environment
from meltano.core.environment_service import EnvironmentService
from meltano.core.manifest import Manifest
from meltano.core.tracking.contexts import CliEvent

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.tracking import Tracker


@click.command(
    "compile",
    cls=InstrumentedCmd,
    short_help="Compile a Meltano manifest. (beta)",
)
@click.option(
    "--directory",
    default=".meltano/manifests",
    help="The path of the directory into which the manifest json files will be written",
    type=click.Path(file_okay=False, resolve_path=True, writable=True, path_type=Path),
)
@click.option(
    "--indent",
    default=4,
    help=(
        "The number of spaces to use for indentation in the manifest JSON "
        "files. Set to -1 to remove all non-essential whitespace."
    ),
)
@click.option(
    # Eventually this lint option will just result in calls to the part of
    # `meltano.core` that backs the `meltano lint` command:
    # https://github.com/meltano/meltano/issues/7285
    "--lint/--no-lint",
    default=False,
    help=(
        "Validate the project files and generated manifest files against the "
        "Meltano schema"
    ),
)
@click.option(
    "--safe/--unsafe",
    default=True,
    show_default=True,
    help="Expose values for sensitive settings.",
)
@click.pass_context
@pass_project(migrate=True)
def compile_command(
    project: Project,
    ctx: click.Context,
    *,
    directory: Path,
    lint: bool,
    indent: int,
    safe: bool,
) -> None:
    """Compile a Meltano project into environment-specific manifest files.

    This command is in beta, and subject to change without corresponding semantic version updates.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#compile
    """  # noqa: D301, E501
    tracker: Tracker = ctx.obj["tracker"]

    try:
        directory.mkdir(parents=True, exist_ok=True)
    except OSError as ex:
        raise CliError(
            f"Unable to create directory for Meltano manifests {directory}: {ex}",  # noqa: EM102
        ) from ex

    environments = _environments(ctx, project)

    click.echo(
        "Compiling Meltano manifest for environments: "
        + ", ".join("no environment" if x is None else x.name for x in environments),
    )
    for environment in environments:
        path = directory / (
            "meltano-manifest.json"
            if environment is None
            else f"meltano-manifest.{environment.name}.json"
        )
        project.refresh(environment=environment)
        manifest = Manifest(
            project=project, path=path, check_schema=lint, redact_secrets=safe
        )
        try:
            with path.open("w") as manifest_file:
                json.dump(
                    manifest.data,
                    manifest_file,
                    indent=indent if indent > 0 else None,
                    sort_keys=True,
                )
        except OSError as ex:
            raise CliError(
                f"Unable to write Meltano manifest {str(path)!r}: {ex}",  # noqa: EM102
            ) from ex

        click.echo(f"Compiled {path}")
        tracker.track_command_event(CliEvent.inflight)


def _environments(
    ctx: click.Context,
    project: Project,
) -> tuple[Environment | None, ...]:
    if ctx.obj["explicit_no_environment"]:
        return (None,)
    if ctx.obj["selected_environment"] and not ctx.obj["is_default_environment"]:
        return (
            Environment.find(
                project.meltano.environments,
                ctx.obj["selected_environment"],
            ),
        )
    return (None, *EnvironmentService(project).list_environments())



"""Add run_id to Job.

Revision ID: 53e97221d99f
Revises: 6ef30ab7b8e5
Create Date: 2019-10-10 13:12:55.147164

"""

from __future__ import annotations

import uuid

import sqlalchemy as sa
import sqlalchemy.orm
from alembic import op

from meltano.migrations import GUID

# revision identifiers, used by Alembic.
revision = "53e97221d99f"
down_revision = "6ef30ab7b8e5"
branch_labels = None
depends_on = None

Session = sa.orm.sessionmaker(future=True)


def upgrade() -> None:
    op.add_column("job", sa.Column("run_id", GUID))

    metadata = sa.MetaData()
    session = Session(bind=op.get_bind())
    Job = sa.Table("job", metadata, autoload_with=op.get_bind())  # noqa: N806

    for job in session.query(Job):
        job.run_id = uuid.uuid4()

    session.commit()


def downgrade() -> None:
    op.drop_column("job", "run_id")



from __future__ import annotations

import os
import shutil

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.plugin import PluginType
from meltano.core.project_add_service import PluginAlreadyAddedException


class TestCliInstall:
    @pytest.fixture(scope="class")
    def tap_gitlab(self, project_add_service):
        try:
            return project_add_service.add(PluginType.EXTRACTORS, "tap-gitlab")
        except PluginAlreadyAddedException as err:
            return err.plugin

    @pytest.mark.order(0)
    def test_install(self, project, tap, tap_gitlab, target, dbt, cli_runner) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(cli, ["install"])
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once_with(
                project,
                [tap, tap_gitlab, target, dbt],
                parallelism=None,
                clean=False,
                force=False,
            )

        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(cli, ["install", "-"])
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once_with(
                project,
                [tap, tap_gitlab, target, dbt],
                parallelism=None,
                clean=False,
                force=False,
            )

    @pytest.mark.usefixtures("dbt")
    def test_install_type(
        self,
        project,
        tap,
        tap_gitlab,
        target,
        mapper,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_e:
            install_plugin_mock_e.return_value = True

            result = cli_runner.invoke(cli, ["install", "extractors"])
            assert_cli_runner(result)

            install_plugin_mock_e.assert_called_once_with(
                project,
                [tap, tap_gitlab],
                parallelism=None,
                clean=False,
                force=False,
            )

        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_l:
            install_plugin_mock_l.return_value = True

            result = cli_runner.invoke(cli, ["install", "loaders"])
            assert_cli_runner(result)

            install_plugin_mock_l.assert_called_once_with(
                project,
                [target],
                parallelism=None,
                clean=False,
                force=False,
            )

        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_m:
            install_plugin_mock_m.return_value = True

            result = cli_runner.invoke(cli, ["install", "mappers"])
            assert_cli_runner(result)

            assert install_plugin_mock_m.call_count == 1
            seen_plugins = install_plugin_mock_m.call_args[0][1]
            assert len(seen_plugins) == 3
            assert mapper in seen_plugins
            mappings_seen = 0
            for found in seen_plugins:
                assert found == mapper
                if found.extra_config.get("_mapping"):
                    mappings_seen += 1
            assert mappings_seen == 2

    @pytest.mark.usefixtures("tap_gitlab", "dbt")
    def test_install_type_name(
        self,
        project,
        tap,
        target,
        mapper,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_e:
            install_plugin_mock_e.return_value = True

            result = cli_runner.invoke(cli, ["install", "extractor", tap.name])
            assert_cli_runner(result)

            install_plugin_mock_e.assert_called_once_with(
                project,
                [tap],
                parallelism=None,
                clean=False,
                force=False,
            )

        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_l:
            install_plugin_mock_l.return_value = True

            result = cli_runner.invoke(cli, ["install", "loader", target.name])
            assert_cli_runner(result)

            install_plugin_mock_l.assert_called_once_with(
                project,
                [target],
                parallelism=None,
                clean=False,
                force=False,
            )

        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock_m:
            install_plugin_mock_m.return_value = True

            result = cli_runner.invoke(cli, ["install", "mapper", mapper.name])
            assert_cli_runner(result)

            assert install_plugin_mock_m.call_count == 1
            seen_plugins = install_plugin_mock_m.call_args[0][1]
            assert len(seen_plugins) == 3
            assert mapper in seen_plugins
            mappings_seen = 0
            for found in seen_plugins:
                assert found == mapper
                if found.extra_config.get("_mapping"):
                    mappings_seen += 1
            assert mappings_seen == 2

    @pytest.mark.usefixtures("target", "dbt")
    def test_install_multiple(self, project, tap, tap_gitlab, cli_runner) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(
                cli,
                ["install", "extractors", tap.name, tap_gitlab.name],
            )
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once_with(
                project,
                [tap, tap_gitlab],
                parallelism=None,
                clean=False,
                force=False,
            )

    @pytest.mark.usefixtures("dbt")
    def test_install_multiple_any_type(
        self,
        project,
        tap,
        target,
        dbt,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(
                cli,
                ["install", "-", tap.name, target.name, dbt.name],
            )
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once_with(
                project,
                [tap, target, dbt],
                parallelism=None,
                clean=False,
                force=False,
            )

    def test_install_parallel(
        self,
        project,
        tap,
        tap_gitlab,
        target,
        dbt,
        mapper,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(cli, ["install", "--parallelism=10"])
            assert_cli_runner(result)

            assert install_plugin_mock.call_count == 1

            commands = [
                args
                for _, args, kwargs in install_plugin_mock.mock_calls
                if args and isinstance(args, tuple)
            ]
            kwargs = install_plugin_mock.mock_calls[0][2]
            assert len(commands) == 1

            assert commands[0][0] == project
            for cli_arg in (dbt, tap_gitlab, target, tap, mapper):
                assert cli_arg in commands[0][1]

            assert kwargs["parallelism"] == 10
            assert not kwargs["clean"]

            mappers = [m for m in commands[0][1] if m == mapper]
            assert len(mappers) == 1

    def test_clean_install(
        self,
        project,
        tap,
        tap_gitlab,
        target,
        dbt,
        mapper,
        cli_runner,
    ) -> None:
        with mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            result = cli_runner.invoke(cli, ["install", "--clean"])
            assert_cli_runner(result)

            assert install_plugin_mock.call_count == 1

            commands = [
                args
                for _, args, kwargs in install_plugin_mock.mock_calls
                if args and isinstance(args, tuple)
            ]
            kwargs = install_plugin_mock.mock_calls[0][2]
            assert len(commands) == 1

            assert commands[0][0] == project
            for cli_arg in (dbt, tap_gitlab, target, tap, mapper):
                assert cli_arg in commands[0][1]

            assert not kwargs["parallelism"]
            assert kwargs["clean"]

            mappers = [m for m in commands[0][1] if m == mapper]
            assert len(mappers) == 1

    @pytest.mark.usefixtures("tap_gitlab", "target")
    def test_install_schedule(
        self,
        project,
        tap_gitlab,
        target,
        dbt,
        mapper,
        cli_runner,
        schedule_service,
        job_schedule,
        task_sets_service,
    ) -> None:
        with mock.patch(
            "meltano.cli.install.ScheduleService",
            return_value=schedule_service,
        ), mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            schedule_service.task_sets_service = task_sets_service
            from meltano.core.task_sets import TaskSets

            mapping = mapper.extra_config.get("_mappings")[0].get("name")
            task_sets_service.add(
                TaskSets(
                    job_schedule.job,
                    [tap_gitlab.name, mapping, target.name, dbt.name],
                ),
            )
            result = cli_runner.invoke(
                cli,
                ["install", "--schedule", job_schedule.name],
            )
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once()
            assert install_plugin_mock.mock_calls[0].args[0] == project

            plugins_installed = [
                plugin.name for plugin in install_plugin_mock.mock_calls[0].args[1]
            ]
            plugins_expected = [tap_gitlab.name, mapper.name, target.name, dbt.name]
            assert sorted(plugins_installed) == sorted(plugins_expected)
            assert install_plugin_mock.mock_calls[0].kwargs["parallelism"] is None
            assert install_plugin_mock.mock_calls[0].kwargs["clean"] is False
            assert install_plugin_mock.mock_calls[0].kwargs["force"] is False

    def test_install_schedule_elt(
        self,
        project,
        tap,
        target,
        cli_runner,
        schedule_service,
        elt_schedule,
        task_sets_service,
    ) -> None:
        with mock.patch(
            "meltano.cli.install.ScheduleService",
            return_value=schedule_service,
        ), mock.patch("meltano.cli.install.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            schedule_service.task_sets_service = task_sets_service

            result = cli_runner.invoke(
                cli,
                ["install", "--schedule", elt_schedule.name],
            )
            assert_cli_runner(result)

            install_plugin_mock.assert_called_once()
            assert install_plugin_mock.mock_calls[0].args[0] == project

            plugins_installed = [
                plugin.name for plugin in install_plugin_mock.mock_calls[0].args[1]
            ]
            plugins_expected = [
                tap.name,
                target.name,
            ]
            assert sorted(plugins_installed) == sorted(plugins_expected)
            assert install_plugin_mock.mock_calls[0].kwargs["parallelism"] is None
            assert install_plugin_mock.mock_calls[0].kwargs["clean"] is False
            assert install_plugin_mock.mock_calls[0].kwargs["force"] is False


# un_engine_uri forces us to create a new project, we must do this before the
# project fixture creates the project see
# https://github.com/meltano/meltano/pull/6407#issuecomment-1200516464
# For more details
@pytest.mark.order(-1)
@pytest.mark.usefixtures("un_engine_uri", "project_function")
def test_new_folder_should_autocreate_on_install(cli_runner) -> None:
    """Be sure .meltano auto creates a db on install by default.

    We had a case https://github.com/meltano/meltano/issues/6383
    that caused the meltano db to not be recreated.
    """
    shutil.rmtree(".meltano")
    result = cli_runner.invoke(cli, ["install"])
    assert result
    assert os.path.exists(".meltano/meltano.db")  # noqa: PTH110



"""Meltano Projects."""

from __future__ import annotations

import errno
import os
import sys
import threading
import typing as t
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path

import fasteners  # type: ignore[import-untyped]
import structlog
from dotenv import dotenv_values

from meltano.core import yaml
from meltano.core.behavior.versioned import Versioned
from meltano.core.config_service import ConfigService
from meltano.core.environment import Environment
from meltano.core.error import (
    EmptyMeltanoFileException,
    ProjectNotFound,
    ProjectReadonly,
)
from meltano.core.hub import MeltanoHubService
from meltano.core.project_files import ProjectFiles
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import makedirs, sanitize_filename, truthy

if t.TYPE_CHECKING:
    from meltano.core.meltano_file import MeltanoFile as MeltanoFileTypeHint
    from meltano.core.plugin.base import PluginRef

    if sys.version_info < (3, 10):
        from typing import TypeAlias  # noqa: ICN003
    else:
        from typing_extensions import TypeAlias


StrPath: TypeAlias = t.Union[str, os.PathLike]

logger = structlog.stdlib.get_logger(__name__)


PROJECT_ROOT_ENV = "MELTANO_PROJECT_ROOT"
PROJECT_ENVIRONMENT_ENV = "MELTANO_ENVIRONMENT"
PROJECT_READONLY_ENV = "MELTANO_PROJECT_READONLY"
PROJECT_SYS_DIR_ROOT_ENV = "MELTANO_SYS_DIR_ROOT"


def walk_parent_directories() -> t.Generator[Path, None, None]:
    """Yield each directory starting with the current up to the root.

    Yields:
        parent directories
    """
    directory = Path.cwd()
    while True:
        yield directory

        parent_directory = directory.parent
        if parent_directory == directory:
            return
        directory = parent_directory


class Project(Versioned):
    """Represents a Meltano project."""

    __version__ = 1
    _activate_lock = threading.Lock()
    _find_lock = threading.Lock()
    _meltano_rw_lock = fasteners.ReaderWriterLock()
    _default = None

    def __init__(
        self,
        root: StrPath,
        environment: Environment | None = None,
        *,
        readonly: bool = False,
    ):
        """Initialize a `Project` instance.

        Args:
            root: The root directory of the project.
            environment: The active Meltano environment.
            readonly: Whether the project is in read-only mode.
        """
        self.root = Path(root).resolve()
        self.environment: Environment | None = environment
        self.readonly = readonly
        self.sys_dir_root = Path(
            os.getenv(PROJECT_SYS_DIR_ROOT_ENV, self.root / ".meltano"),
        ).resolve()

    def refresh(self, **kwargs) -> None:  # noqa: ANN003
        """Refresh the project instance to reflect external changes.

        This should be called whenever env vars change, project files change,
        or other significant changes to the outside world occur.

        Args:
            kwargs: Keyword arguments for the new instance. These overwrite the
                defaults provided by the current instance. For example, if a
                Meltano environment has been activated, the project can be
                refreshed with this new environment by running
                `project.refresh(environment=environment)`.
        """
        kwargs = {
            "root": self.root,
            "environment": self.environment,
            "readonly": self.readonly,
            **kwargs,
        }
        cls = type(self)
        # Clear the dictionary backing `self` to invalidate outdated info,
        # cached properties, etc., then instantiate an up-to-date instance,
        # then steal its attributes to update the dictionary backing `self`.
        # This trick makes it as if the instance was just created, yet keeps
        # all existing references to it valid.
        self.__dict__.clear()
        self.__dict__.update(cls(**kwargs).__dict__)

    @cached_property
    def config_service(self):  # noqa: ANN201
        """Get the project config service.

        Returns:
            A `ConfigService` instance for this project.
        """
        return ConfigService(self)

    @cached_property
    def project_files(self) -> ProjectFiles:
        """Return a singleton `ProjectFiles` file manager instance.

        Returns:
            `ProjectFiles` file manager.
        """
        return ProjectFiles(root=self.root, meltano_file_path=self.meltanofile)

    @cached_property
    def settings(self):  # noqa: ANN201
        """Get the project settings.

        Returns:
            A `ProjectSettingsService` instance for this project.
        """
        return ProjectSettingsService(self)

    @cached_property
    def plugins(self):  # noqa: ANN201
        """Get the project plugins.

        Returns:
            A `ProjectPluginsService` instance for this project.
        """
        return ProjectPluginsService(self)

    @cached_property
    def hub_service(self):  # noqa: ANN201
        """Get the Meltano Hub service.

        Returns:
            A `MeltanoHubService` instance for this project.
        """
        return MeltanoHubService(self)

    @cached_property
    def _meltano_interprocess_lock(self):  # noqa: ANN202
        return fasteners.InterProcessLock(self.run_dir("meltano.yml.lock"))

    @property
    def env(self):  # noqa: ANN201
        """Get environment variables for this project.

        Returns:
            dict of environment variables and values for this project.
        """
        environment_name = self.environment.name if self.environment else ""
        return {
            PROJECT_ROOT_ENV: str(self.root),
            PROJECT_ENVIRONMENT_ENV: environment_name,
            PROJECT_SYS_DIR_ROOT_ENV: str(self.sys_dir_root),
        }

    @classmethod
    @fasteners.locked(lock="_activate_lock")
    def activate(cls, project: Project) -> None:
        """Activate the given Project.

        Args:
            project: the Project to activate

        Raises:
            OSError: if project cannot be activated due to unsupported OS
        """
        import ctypes

        project.ensure_compatible()

        # create a symlink to our current binary
        try:
            # check if running on Windows
            if os.name == "nt":
                executable = Path(sys.executable).parent / "meltano.exe"
                # Admin privileges are required to create symlinks on Windows
                if ctypes.windll.shell32.IsUserAnAdmin():  # type: ignore[attr-defined]
                    if executable.is_file():
                        project.run_dir().joinpath("bin").symlink_to(executable)
                    else:
                        logger.warning(
                            "Could not create symlink: meltano.exe not "  # noqa: G004
                            f"present in {Path(sys.executable).parent!s}",
                        )
                else:
                    logger.warning(
                        "Failed to create symlink to 'meltano.exe': "
                        "administrator privilege required",
                    )
            else:
                executable = Path(sys.executable).parent / "meltano"
                if executable.is_file():
                    project.run_dir().joinpath("bin").symlink_to(executable)
        except FileExistsError:
            pass
        except OSError as error:
            if error.errno == errno.EOPNOTSUPP:
                logger.warning(
                    f"Could not create symlink: {error}\nPlease make sure "  # noqa: G004
                    "that the underlying filesystem supports symlinks.",
                )
            else:
                raise

        logger.debug("Activated project at %s", project.root)

        # set the default project
        cls._default = project

    @classmethod
    def deactivate(cls) -> None:
        """Deactivate the given Project."""
        cls._default = None

    @property
    def file_version(self):  # noqa: ANN201
        """Get the version of Meltano found in this project's meltano.yml.

        Returns:
            the Project's meltano version
        """
        return self.meltano.version

    @classmethod
    @fasteners.locked(lock="_find_lock")
    def find(cls, project_root: Path | str | None = None, *, activate=True):  # noqa: ANN001, ANN206
        """Find a Project.

        Args:
            project_root: The path to the root directory of the project. If not
                supplied, infer from PROJECT_ROOT_ENV or the current working
                directory and it's parents.
            activate: Save the found project so that future calls to `find`
                will continue to use this project.

        Returns:
            the found project

        Raises:
            ProjectNotFound: if the provided `project_root` is not a Meltano
                project, or the current working directory is not a Meltano
                project or a subfolder of one.
        """
        if cls._default:
            return cls._default

        readonly = truthy(os.getenv(PROJECT_READONLY_ENV, "false"))

        if project_root := project_root or os.getenv(PROJECT_ROOT_ENV):
            project = Project(project_root, readonly=readonly)
            if not project.meltanofile.exists():
                raise ProjectNotFound(project)
        else:
            for directory in walk_parent_directories():
                project = Project(directory, readonly=readonly)
                if project.meltanofile.exists():
                    break
            if not project.meltanofile.exists():
                raise ProjectNotFound(Project(Path.cwd()))

        readonly = project.settings.get("project_readonly")
        if readonly != project.readonly:
            project.refresh(readonly=readonly)

        if activate:
            cls.activate(project)

        return project

    @property
    def meltano(self) -> MeltanoFileTypeHint:
        """Return a copy of the current meltano config.

        Raises:
            EmptyMeltanoFileException: The `meltano.yml` file is empty.

        Returns:
            The current meltano config.
        """
        from meltano.core.meltano_file import MeltanoFile

        conf: dict[str, t.Any] = yaml.load(self.meltanofile)
        if conf is None:
            raise EmptyMeltanoFileException

        with self._meltano_rw_lock.read_lock():
            return MeltanoFile.parse(self.project_files.load())

    @contextmanager
    def meltano_update(self):  # noqa: ANN201
        """Yield the current meltano configuration.

        Update the meltanofile if the context ends gracefully.

        Yields:
            the current meltano configuration

        Raises:
            ProjectReadonly: This project is readonly.
            Exception: The project files could not be updated.
        """
        if self.readonly:
            raise ProjectReadonly

        from meltano.core.meltano_file import MeltanoFile

        with self._meltano_rw_lock.write_lock(), self._meltano_interprocess_lock:
            meltano_config = MeltanoFile.parse(self.project_files.load())
            yield meltano_config
            try:
                self.project_files.update(meltano_config.canonical())
            except Exception as err:
                logger.critical("Could not update meltano.yml: %s", err)
                raise

        self.refresh()

    def root_dir(self, *joinpaths: StrPath) -> Path:
        """Return the root directory of this project, optionally joined with path.

        Args:
            joinpaths: list of subdirs and/or file to join to project root.

        Returns:
            project root joined with provided subdirs and/or file
        """
        return self.root.joinpath(*joinpaths)

    @property
    def meltanofile(self) -> Path:
        """Get the path to this project's meltano.yml.

        Returns:
            the path to this project meltano.yml
        """
        return self.root.joinpath("meltano.yml")

    @property
    def dotenv(self) -> Path:
        """Get the path to this project's .env file.

        Returns:
            the path to this project's .env file
        """
        return self.root.joinpath(".env")

    @cached_property
    def dotenv_env(self) -> dict[str, str | None]:
        """Get values from this project's .env file.

        Returns:
            values found in this project's .env file
        """
        return dotenv_values(self.dotenv)

    def activate_environment(self, name: str) -> None:
        """Activate a Meltano environment.

        No-op if the active environment has the given name.

        Args:
            name: Name of the environment.
        """
        if getattr(self.environment, "name", object()) != name:
            self.refresh(environment=Environment.find(self.meltano.environments, name))
        logger.info(f"Environment {name!r} is active")  # noqa: G004

    def deactivate_environment(self) -> None:
        """Deactivate the currently active environment."""
        if self.environment is not None:
            self.refresh(environment=None)

    @contextmanager
    def dotenv_update(self):  # noqa: ANN201
        """Raise error if project is readonly.

        Used in context where .env files would be updated.

        Yields:
            the .env file

        Raises:
            ProjectReadonly: if the project is readonly
        """
        if self.readonly:
            raise ProjectReadonly

        yield self.dotenv
        self.refresh()

    @makedirs
    def meltano_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:  # noqa: ARG002
        """Path to the project `.meltano` directory.

        Args:
            joinpaths: Paths to join to the `.meltano` directory.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `.meltano` dir optionally joined to given paths.
        """
        return self.sys_dir_root.joinpath(*joinpaths)

    @makedirs
    def analyze_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:  # noqa: ARG002
        """Path to the project `analyze` directory.

        Args:
            joinpaths: Paths to join to the `analyze` directory.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `analyze` dir optionally joined to given paths.
        """
        return self.root_dir("analyze", *joinpaths)

    @makedirs
    def extract_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:  # noqa: ARG002
        """Path to the project `extract` directory.

        Args:
            joinpaths: Paths to join to the `extract` directory.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `extract` dir optionally joined to given paths.
        """
        return self.root_dir("extract", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes: StrPath, make_dirs: bool = True) -> Path:
        """Path to a `venv` directory in `.meltano`.

        Args:
            prefixes: Paths to prepend to the `venv` directory in `.meltano`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `venv` dir optionally prepended with given prefixes.
        """
        return self.meltano_dir(*prefixes, "venv", make_dirs=make_dirs)

    @makedirs
    def run_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:
        """Path to the `run` directory in `.meltano`.

        Args:
            joinpaths: Paths to join to the `run` directory in `.meltano`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `run` dir optionally joined to given paths.
        """
        return self.meltano_dir("run", *joinpaths, make_dirs=make_dirs)

    @makedirs
    def logs_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:
        """Path to the `logs` directory in `.meltano`.

        Args:
            joinpaths: Paths to join to the `logs` directory in `.meltano`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `logs` dir optionally joined to given paths.
        """
        return self.meltano_dir("logs", *joinpaths, make_dirs=make_dirs)

    @makedirs
    def job_dir(self, state_id, *joinpaths: StrPath, make_dirs: bool = True) -> Path:  # noqa: ANN001
        """Path to the `elt` directory in `.meltano/run`.

        Args:
            state_id: State ID of `run` dir.
            joinpaths: Paths to join to the `elt` directory in `.meltano`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `elt` dir optionally joined to given paths.
        """
        return self.run_dir(
            "elt",
            sanitize_filename(state_id),
            *joinpaths,
            make_dirs=make_dirs,
        )

    @makedirs
    def job_logs_dir(
        self,
        state_id,  # noqa: ANN001
        *joinpaths: StrPath,
        make_dirs: bool = True,
    ) -> Path:
        """Path to the `elt` directory in `.meltano/logs`.

        Args:
            state_id: State ID of `logs` dir.
            joinpaths: Paths to join to the `elt` directory in `.meltano/logs`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to `elt` dir optionally joined to given paths.
        """
        return self.logs_dir(
            "elt",
            sanitize_filename(state_id),
            *joinpaths,
            make_dirs=make_dirs,
        )

    @makedirs
    def plugin_dir(
        self,
        plugin: PluginRef,
        *joinpaths: StrPath,
        make_dirs: bool = True,
    ) -> Path:
        """Path to the plugin installation directory in `.meltano`.

        Args:
            plugin: Plugin to retrieve or create directory for.
            joinpaths: Paths to join to the plugin installation directory in `.meltano`.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Resolved path to plugin installation dir optionally joined to given paths.
        """
        return self.meltano_dir(
            plugin.type,
            plugin.plugin_dir_name,
            *joinpaths,
            make_dirs=make_dirs,
        )

    @makedirs
    def root_plugins_dir(self, *joinpaths: StrPath, make_dirs: bool = True) -> Path:  # noqa: ARG002
        """Path to the project `plugins` directory.

        Args:
            joinpaths: Paths to join with the project `plugins` directory.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Path to the project `plugins` directory.
        """
        return self.root_dir("plugins", *joinpaths)

    @makedirs
    def plugin_lock_path(  # noqa: ANN201
        self,
        plugin_type: str,
        plugin_name: str,
        *,
        variant_name: str | None = None,
        make_dirs: bool = True,
    ):
        """Path to the project lock file.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.
            make_dirs: Whether to create the directory hierarchy if it doesn't exist.

        Returns:
            Path to the plugin lock file.
        """
        filename = f"{plugin_name}"

        if variant_name:
            filename = f"{filename}--{variant_name}"

        return self.root_plugins_dir(
            plugin_type,
            f"{filename}.lock",
            make_dirs=make_dirs,
        )

    def __eq__(self, other):  # noqa: ANN001, ANN204
        """Project equivalence check.

        Args:
            other: The other Project instance to check against.

        Returns:
            True if Projects are equal.
        """
        return self.root == getattr(other, "root", object())

    def __hash__(self):  # noqa: ANN204
        """Project hash.

        Returns:
            Project hash.
        """
        return self.root.__hash__()



from __future__ import annotations

import typing as t

import mock
import pytest
from mock import AsyncMock

from meltano.core.job import Job, Payload, State
from meltano.core.logging.utils import capture_subprocess_output
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.runner.singer import SingerRunner

if t.TYPE_CHECKING:
    from pathlib import Path

    from meltano.core.plugin.project_plugin import ProjectPlugin

TEST_STATE_ID = "test_job"


class AnyInstanceOf:
    def __init__(self, target_cls) -> None:
        self.target_cls = target_cls

    def __eq__(self, other):
        return isinstance(other, self.target_cls)

    def __repr__(self) -> str:
        return f"<Any({self.target_cls}>"


def create_plugin_files(config_dir: Path, plugin: ProjectPlugin):
    for file in plugin.config_files.values():
        config_dir.joinpath(file).touch()

    return config_dir


class TestSingerRunner:
    @pytest.fixture()
    def elt_context(
        self,
        project,  # noqa: ARG002
        session,
        tap,
        target,
        elt_context_builder,
    ):
        job = Job(job_name="pytest_test_runner")

        return (
            elt_context_builder.with_session(session)
            .with_extractor(tap.name)
            .with_job(job)
            .with_loader(target.name)
            .context()
        )

    @pytest.fixture()
    def tap_config_dir(self, tmp_path: Path, elt_context) -> Path:
        create_plugin_files(tmp_path, elt_context.extractor.plugin)
        return tmp_path

    @pytest.fixture()
    def target_config_dir(self, tmp_path: Path, elt_context) -> Path:
        create_plugin_files(tmp_path, elt_context.loader.plugin)
        return tmp_path

    @pytest.fixture()
    def subject(self, session, elt_context):
        Job(
            job_name=TEST_STATE_ID,
            state=State.SUCCESS,
            payload_flags=Payload.STATE,
            payload={"singer_state": {"bookmarks": []}},
        ).save(session)

        return SingerRunner(elt_context)

    @pytest.fixture()
    def process_mock_factory(self):
        def _factory(name):
            process_mock = mock.Mock()
            process_mock.name = name
            process_mock.wait = AsyncMock(return_value=0)
            return process_mock

        return _factory

    @pytest.fixture()
    def tap_process(self, process_mock_factory, tap):
        tap = process_mock_factory(tap)
        tap.stdout.readline = AsyncMock(return_value="{}")
        return tap

    @pytest.fixture()
    def target_process(self, process_mock_factory, target):
        return process_mock_factory(target)

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("subject")
    async def test_prepare_job(
        self,
        session,
        tap_config_dir,
        target_config_dir,
        tap,
        target,
        plugin_invoker_factory,
    ) -> None:
        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)

        async with tap_invoker.prepared(session):
            for name in tap.config_files:
                assert tap_invoker.files[name].exists()

        assert not tap_invoker.files["config"].exists()

        async with target_invoker.prepared(session):
            for name in target.config_files:
                assert target_invoker.files[name].exists()

        assert not target_invoker.files["config"].exists()

    @pytest.mark.asyncio()
    async def test_invoke(
        self,
        session,
        subject,
        tap_config_dir,
        target_config_dir,
        tap,
        target,
        tap_process,
        target_process,
        plugin_invoker_factory,
    ) -> None:
        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)
        async with tap_invoker.prepared(
            session,
        ), target_invoker.prepared(session):
            invoke_async = AsyncMock(side_effect=(tap_process, target_process))
            with mock.patch.object(
                PluginInvoker,
                "invoke_async",
                new=invoke_async,
            ) as invoke_async:
                # async method
                await subject.invoke(tap_invoker, target_invoker, session)

                # correct bins are called
                assert await invoke_async.awaited_with(tap_invoker)
                assert await invoke_async.awaited_with(target_invoker)

                tap_process.wait.assert_awaited()
                target_process.wait.assert_awaited()

    @pytest.mark.asyncio()
    @pytest.mark.parametrize(
        ("full_refresh", "merge_state", "select_filter", "payload_flag"),
        (
            pytest.param(
                False,
                False,
                [],
                Payload.STATE,
                id="incremental-no-merge-no-select--complete-state",
            ),
            pytest.param(
                True,
                False,
                [],
                Payload.STATE,
                id="full-refresh-no-merge-no-select--complete-state",
            ),
            pytest.param(
                False,
                False,
                ["entity"],
                Payload.STATE,
                id="incremental-no-merge-select--complete-state",
            ),
            pytest.param(
                True,
                False,
                ["entity"],
                Payload.INCOMPLETE_STATE,
                id="full-refresh-no-merge-select--incomplete-state",
            ),
            pytest.param(
                False,
                True,
                [],
                Payload.INCOMPLETE_STATE,
                id="incremental-merge-no-select--incomplete-state",
            ),
            pytest.param(
                True,
                True,
                [],
                Payload.INCOMPLETE_STATE,
                id="full-refresh-merge-no-select--incomplete-state",
            ),
            pytest.param(
                False,
                True,
                ["entity"],
                Payload.INCOMPLETE_STATE,
                id="incremental-merge-select--incomplete-state",
            ),
            pytest.param(
                True,
                True,
                ["entity"],
                Payload.INCOMPLETE_STATE,
                id="full-refresh-merge-select--incomplete-state",
            ),
        ),
    )
    async def test_bookmark(
        self,
        subject,
        session,
        target,
        target_config_dir,
        target_process,
        plugin_invoker_factory,
        full_refresh,
        select_filter,
        payload_flag,
        elt_context,
        merge_state,
    ) -> None:
        lines = (b'{"line": 1}\n', b'{"line": 2}\n', b'{"line": 3}\n')

        # testing with a real subprocess proved to be pretty
        # complicated.
        target_process.stdout = mock.MagicMock()
        target_process.stdout.at_eof.side_effect = (False, False, False, True)
        target_process.stdout.readline = AsyncMock(side_effect=lines)

        subject.context.full_refresh = full_refresh
        subject.context.select_filter = select_filter
        subject.context.merge_state = merge_state

        target_invoker = plugin_invoker_factory(
            target,
            config_dir=target_config_dir,
            context=elt_context,
        )

        async with subject.context.job.run(session):
            with mock.patch.object(
                session,
                "add",
                side_effect=session.add,
            ) as add_mock, mock.patch.object(
                session,
                "commit",
                side_effect=session.commit,
            ) as commit_mock:
                target.setup_bookmark_writer(target_invoker)
                bookmark_writer = target_invoker.output_handlers.get(
                    target_invoker.StdioSource.STDOUT,
                )
                await capture_subprocess_output(target_process.stdout, *bookmark_writer)

            assert add_mock.call_count == 9
            assert commit_mock.call_count == 9

            # assert the STATE's `value` was saved
            job = subject.context.job
            assert job.payload["singer_state"] == {"line": 3}
            assert job.payload_flags == payload_flag

    @pytest.mark.asyncio()
    async def test_run(self, subject) -> None:
        async def invoke_mock(*args, **kwargs) -> None:
            pass

        with mock.patch.object(
            SingerRunner,
            "invoke",
            side_effect=invoke_mock,
        ) as invoke:
            subject.context.dry_run = True
            await subject.run()

            assert not invoke.called

            subject.context.dry_run = False
            await subject.run()
            AnyPluginInvoker = AnyInstanceOf(PluginInvoker)  # noqa: N806

            invoke.assert_called_once_with(
                AnyPluginInvoker,
                AnyPluginInvoker,
                extractor_log=None,
                loader_log=None,
                extractor_out=None,
                loader_out=None,
            )
